#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import os
from SMART.Java.Python.getSizes import GetSizes
from SMART.Java.Python.clusterize import Clusterize
from SMART.Java.Python.ComputeCoverage import CoverageComputer
from commons.core.seq.BioseqDB import BioseqDB
from commons.core.LoggerFactory import LoggerFactory
from commons.launcher.LaunchTallymer import LaunchTallymer
from commons.core.utils.RepetOptionParser import RepetOptionParser

LOG_DEPTH = "repet.Tallymer_pipe"

class TallymerPipe(object):
    
    def __init__(self, inputFastaFileName = None, indexationFasta = None, inputReferenceAnnot = None, nLargestScaffoldsToPlot = 0, steps="1,2,3", mersize = 20, minOccs = 4,\
                 clusterizeDistance = 1, windowSize = 100000, windowOverlap = 10000, outputFormats = "gff", verbosity = 3):
        
        self._inFastaFileName = inputFastaFileName
        self._indexationFasta = indexationFasta if indexationFasta != None else inputFastaFileName 
        self._inputReferenceAnnot = inputReferenceAnnot
        self._nLargestScaffoldsToPlot = nLargestScaffoldsToPlot
        self._outputFormats = outputFormats 
        self._steps = steps
        self._merSize = mersize
        self._minOccs = minOccs
        self._clusterizeDistance = clusterizeDistance
        self.windowSize = windowSize
        self.windowOverlap = windowOverlap
        self._workdir = os.getcwd()
        
        self.doClean = False
        self._verbosity = verbosity
        self._subscriptVerbosity = 0 if self._verbosity<4 else self._verbosity-1
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbosity)
    
    def setAttributesFromCmdLine(self):
        description = "."    
        parser = RepetOptionParser(description = description)
        parser.add_option("-i", "--inputFasta",         dest="inputFasta",          default = None,   action="store", type="string", help="input fasta file [compulsory] [format: fasta]")
        parser.add_option("-r", "--inputReferenceAnnot",dest="inputReferenceAnnot", default = None,   action="store", type="string", help="reference annotation file [optional] [format: gff]")
        parser.add_option("-u", "--indexationFasta",    dest="indexationFasta",     default = None,   action="store", type="string", help="input indexation fasta file used to generate kmer index (defaults to input fasta)[optional] [format: fasta]")
        parser.add_option("-f", "--outputFormats",      dest="outputFormats",       default = "gff",  action="store", type="string", help="comma separated list of output file formats [optional] [default: gff]")
        parser.add_option("-m", "--merSize",            dest="merSize",             default = 20,     action="store", type="int",    help="input mer size (K-value) [optional] [default: 20]")
        parser.add_option("-z", "--minOccs",            dest="minOccs",             default = 4,      action="store", type="int",    help="input minimal K-mer occurrences count [default: 4]")
        parser.add_option("-d", "--distance",           dest="distance",            default = 1,      action="store", type="int",    help="max. distance between two K-mer occurrences to be merged [default: 1]")
        parser.add_option("-w", "--windowSize",         dest="windowSize",          default = 100000, action="store", type="int",    help="sliding window size [optional] [default: 100000]")
        parser.add_option("-y", "--windowOverlap",      dest="windowOverlap",       default = 10000,  action="store", type="int",    help="sliding window overlap size [optional] [default: 10000]")
        # parser.add_option("-n", "--nLargestScaffoldsToPlot",default = 0, type="int", action="store", dest = "nLargestScaffoldsToPlot", help = "generate graph of Kmer repartition along the input sequence for the n biggest scaffolds")
        parser.add_option("-S", "--steps",              dest="steps",               default = "123",  action="store", type="string", help="steps to run [optional] [default: 123]")
        parser.add_option("-c", "--clean",              dest="clean",               default = False,  action="store_true",           help="clean temporary files")
        parser.add_option("-v", "--verbosity",          dest="verbosity",           default = 3,      action="store", type="int",    help="verbosity [optional] [default: 3]")
        options, args = parser.parse_args()
        self._setAttributesFromOptions(options)
        
    def _setAttributesFromOptions(self, options):
        self._inFastaFileName = options.inputFasta
        self._indexationFasta = options.indexationFasta if options.indexationFasta != None else options.inputFasta
        self._inputReferenceAnnot = options.inputReferenceAnnot
        self._outputFormats = options.outputFormats
        #self._nLargestScaffoldsToPlot = options.nLargestScaffoldsToPlot
        self._steps = options.steps
        self._merSize = options.merSize
        self._minOccs = options.minOccs
        self._clusterizeDistance = options.distance
        self.windowSize = options.windowSize
        self.windowOverlap = options.windowOverlap
        self.doClean = options.clean
        self._verbosity = options.verbosity
        self._subscriptVerbosity = 0 if self._verbosity< 4 else self._verbosity-1
   
    def run(self): 
        LoggerFactory.setLevel(self._log, self._verbosity)
        self._log.info("Starting TallymerPipe with input fasta : %s, tallymer index fasta %s , reference file:%s" %(self._inFastaFileName, self._indexationFasta, self._inputReferenceAnnot))
        self._outputFormats = self._outputFormats + ",gff" if not "gff" in self._outputFormats else self._outputFormats
        self._outputFormats = self._outputFormats.replace(' ', '')

        try:
            os.makedirs(self._workdir)
        except:pass
   
        srcPath = os.path.abspath(self._inFastaFileName)
        dstPath = os.path.join(self._workdir, os.path.basename(self._inFastaFileName))
        try:
            os.symlink(srcPath, dstPath)
        except: pass
        inputFastaFileName = os.path.join(self._workdir, os.path.basename(srcPath))
        
        if self._inputReferenceAnnot is not None:
            srcPath = os.path.abspath(self._inputReferenceAnnot)
            dstPath = os.path.join(self._workdir, os.path.basename(self._inputReferenceAnnot))
            try:
                os.symlink(srcPath, dstPath)
            except:pass
            inputReferenceFileName = os.path.join(self._workdir, os.path.basename(srcPath))
        
        if self._indexationFasta != self._inFastaFileName:
            srcPath = os.path.abspath(self._indexationFasta)
            dstPath = os.path.join(self._workdir, os.path.basename(self._indexationFasta))
            try:
                os.symlink(srcPath, dstPath)
            except:pass
            inputIndexationFileName = os.path.join(self._workdir, os.path.basename(srcPath))
        else:
            inputIndexationFileName = self._inFastaFileName
            
        os.chdir(self._workdir)
        format = "gff"
        self.stats = {}
       
        srcFileName = os.path.basename(self._inFastaFileName)
        
        project_name = os.path.splitext(os.path.basename(self._inFastaFileName))[0].replace(" ","_")
        step1_dir = "TallymerPipe_Tallymer_%s" % project_name 
        step2_dir = "TallymerPipe_ClusterizeAndSlidingWindow_%s" % project_name
        step3_dir = "TallymerPipe_CoverageAndStats_%s" % project_name
        step4_dir = "TallymerPipe_Plots_%s" % project_name
        
        tallymerOutputFileName = os.path.join(self._workdir, step1_dir, "%s.tallymer.%s" % (os.path.splitext(srcFileName)[0], format))
        clusterizedFileName = os.path.join(self._workdir, step2_dir, "%s.tallymer_clusterized.%s" % (os.path.splitext(srcFileName)[0], format))
        clusterizedWindowsFileName = os.path.join(self._workdir, step2_dir, "%s.tallymer_clusterized_windows.%s" % (os.path.splitext(srcFileName)[0], format))
        coverageClusterizedWindowsFileName = os.path.join(self._workdir, step3_dir, "%s.tallymer_clusterized_windows_coverage.%s" % (os.path.splitext(srcFileName)[0], format))
        
        if "1" in self._steps:
            self._log.info("Starting Tallymer on %s" % self._inFastaFileName)
            stepDir = os.path.join(self._workdir, step1_dir)
            try:
                os.makedirs(stepDir)
            except:pass
            os.chdir(stepDir)
            iLaunchTallymer = LaunchTallymer(inputFastaFileName, inputIndexationFileName, merSize = self._merSize,minOccs = self._minOccs, outputFormats = self._outputFormats,\
                                             nLargestScaffoldsToPlot = 0, clean = self.doClean, verbosity = self._subscriptVerbosity)
            iLaunchTallymer.run()
            self._log.info("Finished running Tallymer")
            os.chdir("..")
            
        if "2" in self._steps: 
            stepDir = os.path.join(self._workdir, step2_dir)
            try:
                os.makedirs(stepDir)
            except:pass
            os.chdir(stepDir)
            self._log.info("Starting clusterize on %s" % tallymerOutputFileName)
            c = Clusterize(self._subscriptVerbosity)
            c.setInputFile(tallymerOutputFileName, format)
            c.setOutputFileName(clusterizedFileName, format, title = "TallymerPipe", feature = "Match", featurePart = "Match-Part")
            c.setPresorted(True)
            c.setDistance(self._clusterizeDistance)
            c.run()
            self._log.info("clusterize done")
            
            self._log.info("Doing sliding window: size: %d, overlap %d" % (self.windowSize, self.windowOverlap))
#            cbsw = ClusterizeBySlidingWindows(inputFileNames=clusterizedFileName,inputFormat=format,defaultValue=1, tag="nbElements", outputTagNames="nbElements", 
#                                              size =self.windowSize, overlap = self.windowOverlap, operation="sum", verbosity=self._subscriptVerbosity)
#            print("clusterizedFileName",clusterizedFileName)
#            cbsw.setWindowSize(self.windowSize)
#            cbsw.setWindowOverlap(self.windowOverlap)
#            cbsw.setTag("nbElements")
#            cbsw.setOperation("sum")
#            cbsw.setOutputTag("nbElements")
#            cbsw.setDefaultValue(1)
#            cbsw.setInputFile(clusterizedFileName, format)
#            cbsw.setOutputFileName(clusterizedWindowsFileName,format,title="TallymerPipe", feature="Match", featurePart="Match-Part")
#            cbsw.run()

#Error: missing window 2!!
#Chr1    TallymerPipe    Match    1    100000    .    +    .    nbElements=548.000000;ID=region1;Name=region1
#Chr1    TallymerPipe    Match    180001    280000    .    +    .    nbElements=197.000000;ID=region2;Name=region2
#Chr1    TallymerPipe    Match    270001    370000    .    +    .    nbElements=45.000000;ID=region3;Name=region3

            
            cmd = "python %s/SMART/Java/Python/clusterizeBySlidingWindows.py" % os.environ["REPET_PATH"]
            cmd += " -i %s" % clusterizedFileName
            cmd += " -o %s" % clusterizedWindowsFileName
            cmd += " -f gff3"
            cmd += " -s %d" % self.windowSize
            cmd += " -e %d" % self.windowOverlap
            cmd += " -v %d" % self._subscriptVerbosity
            cmd += " -z %s" % "TallymerPipe,Match,Match-Part"
            os.system(cmd)
            
            self._log.info("sliding window sum done")
            os.chdir("..")

        if "3" in self._steps:
            stepDir = os.path.join(self._workdir,step3_dir)
            try:
                os.makedirs(stepDir)
            except:pass
            os.chdir(stepDir)
            
            self._log.info("Computing coverage clusterized_sliding_window vs clusterized")
            self._compute_coverage(clusterizedWindowsFileName, format, clusterizedFileName, format, coverageClusterizedWindowsFileName)
            self._log.info("Finished computing coverage clusterized_sliding_window vs clusterized")
            
            if self._inputReferenceAnnot is not None:
                self._log.info("Computing coverage clusterized_sliding_window vs annotation")
                self._compute_coverage(clusterizedWindowsFileName, format, inputReferenceFileName, format, coverageClusterizedWindowsFileName + "annot")
                self._log.info("Finished computing coverage clusterized_sliding_window vs annotation")
        
            self._log.info("Generating coverage/sizes stats")
            iGetSizes = GetSizes(inFileName = clusterizedFileName, inFormat = format)
            try:
                iGetSizes.run()
            except:
                pass
            
            self.stats["clusterized file info"] = "No result"
            self.stats["reference file info"] = "No reference provided"
            self.stats["input fasta file info"] = "No input fasta provided"
            self.stats["coverage % of input fasta"] = "Could no compute coverage"
            self.stats["ratio (coverage clusterized file)/(coverage reference file)"] = "Could no compute coverage"
            
            self.stats["clusterized file info"] = "Items: %d subitems: %d\n Nucleotides: %d\n Min/avg/med/max transcripts: %d/%.2f/%.1f/%d\n" % (iGetSizes.items,\
                      iGetSizes.subItems, iGetSizes.nucleotides, iGetSizes.minAvgMedMax[0], iGetSizes.minAvgMedMax[1], iGetSizes.minAvgMedMax[2], iGetSizes.minAvgMedMax[3])
            
            if self._inputReferenceAnnot != None:
                iGetSizes2 = GetSizes(inFileName = inputReferenceFileName, inFormat = format)
                iGetSizes2.run()
                self.stats["reference file info"] = "Items: %d subitems: %d\n Nucleotides: %d\n Min/avg/med/max transcripts: %d/%.2f/%.1f/%d\n" % (iGetSizes2.items,\
                          iGetSizes2.subItems, iGetSizes2.nucleotides, iGetSizes2.minAvgMedMax[0], iGetSizes2.minAvgMedMax[1], iGetSizes2.minAvgMedMax[2], iGetSizes2.minAvgMedMax[3])
                self.stats["ratio (coverage clusterized file)/(coverage reference file)"] = (float(iGetSizes.nucleotides) / float(iGetSizes2.nucleotides)) * 100
                
            if self._inFastaFileName != None:
                iGetSizes3 = GetSizes(inFileName = inputFastaFileName, inFormat = "fasta")
                iGetSizes3.run()
                self.stats["input fasta file info"] = "Items: %d subitems: %d\n Nucleotides: %d\n Min/avg/med/max transcripts: %d/%.2f/%.1f/%d\n" % (iGetSizes3.items,\
                          iGetSizes3.subItems, iGetSizes3.nucleotides, iGetSizes3.minAvgMedMax[0], iGetSizes3.minAvgMedMax[1], iGetSizes3.minAvgMedMax[2], iGetSizes3.minAvgMedMax[3])
                self.stats["coverage % of input fasta"] = (float(iGetSizes.nucleotides) / float(iGetSizes3.nucleotides)) * 100    
                
            with open("tallymerPipe.stats",'w') as tallymerPipeStatsF:
                tallymerPipeStatsF.write("===========TallymerPipe outputs================\n")
                tallymerPipeStatsF.write("TallymerOutput File: %s \nReference OutputFile: %s\n" % (clusterizedFileName, self._inputReferenceAnnot))
                tallymerPipeStatsF.write("===========clusterized file info===========\n %s \n" % self.stats["clusterized file info"])
                tallymerPipeStatsF.write("===========Reference file info===========\n %s \n" % self.stats["reference file info"])
                tallymerPipeStatsF.write("===========Input fasta file info===========\n %s \n" % self.stats["input fasta file info"])
                tallymerPipeStatsF.write("Ratio (coverage clusterized file)/(coverage reference file): %s \n" % self.stats["ratio (coverage clusterized file)/(coverage reference file)"])
                tallymerPipeStatsF.write("Total coverage percent of input fasta: %s \n" % self.stats["coverage % of input fasta"])

            self._log.info("Finished Generating coverage stats")
            os.chdir("..")
        
         
#        if "4" in self._steps:
#            stepDir = os.path.join(self._workdir,step4_dir)
#            try:
#                os.makedirs(stepDir)
#            except:pass
#            os.chdir(stepDir)
#            self._log.info("Generating plots")
#            if self._nLargestScaffoldsToPlot>0:
#                self.generate_simple_plot(inputFastaFileName, os.path.abspath(coverageClusterizedWindowsFileName))
#                try:
#                    self.generate_circos_plot(inputFastaFileName, coverageClusterizedWindowsFileName)
#                except Exception as inst:
#                    self._log.error("Failed to generate circos plot because of error %s" % inst)
#            self._log.info("Finished generating plots")
#            os.chdir("..")
            
    def generate_simple_plot(self, inputFastaFileName, inFile):
        if self._nLargestScaffoldsToPlot > 0:
            self._log.info("Generating simple plot for the  %d largest scaffolds/chr" % self._nLargestScaffoldsToPlot)
            iBioseqDB = BioseqDB()
            iBioseqDB.load(inputFastaFileName)
            largest_seqsDb = iBioseqDB.bestLength(self._nLargestScaffoldsToPlot)
            
            for seq in largest_seqsDb.db:
                header = seq.header.replace(" ","_")
                cmd = "grep %s %s > %s.gff" % (header, inFile, header)
                os.system(cmd)
                subFile = "%s.gff" % header
                
                cmd = "Rscript %s/Dojo/TallymerPipe/tallymer_plot.r %s %s %s.png %d %d " % (os.environ["REPET_PATH"], subFile, header, '"Coverage "', 2000, 1000)
                os.system(cmd)
            
            self._log.info("simple plot done")
        
    def generate_circos_plot(self, inFile, inClusterizedFile):
        format = "gff"
        iBioseqDB = BioseqDB()
        iBioseqDB.load(inFile)
        largest_seqsDb = iBioseqDB.bestLength(self._nLargestScaffoldsToPlot)
        
        with open("circosChr.txt",'w') as outFileHandle:
            color = "black"
            for seq in largest_seqsDb.db:
                outFileHandle.write("chr - %s\t%s\t0\t%s\t%s\n" % (seq.header, seq.header, iBioseqDB.getSeqLength(seq.header), color))
                
        #tallymerOutputFileName = "%s.tallymer_filtered_clusterized_filtered_sliding.%s" % (os.path.splitext(os.path.basename(self._inFastaFileName))[0], format)
        #convert data format to circos format:
        with open("circosInput.txt",'w') as outFileHandle:
            with open(inClusterizedFile) as inFileHandle:
                for index, line in enumerate(inFileHandle):
                    lineElems = line.split("\t")
                    chr = lineElems[0]
                    start = lineElems[3]
                    end = lineElems[4] 
                    sumNbelements = lineElems[8].split(";")[2].split("=")[1]
                    #TODO : this is wrong : there is a karyotype input and a data input file , which are different
                    rstart = (float(end)-float(start))/2
                    rend = (float(end)-float(start))/2
                    #outFileHandle.write("%s\t%d\t%d\t%s\n"%(chr,rstart,rend,sumNbelements))
                    outFileHandle.write("%s\t%s\t%s\t%s\n" % (chr, start, end, sumNbelements))
        
        #for annotation
        with open("circosInputRef.txt", 'w') as outFileHandle:
            with open(inClusterizedFile) as inFileHandle:
            #with open(self._inputReferenceAnnot) as inFileHandle:
                for index, line in enumerate(inFileHandle):
                    lineElems = line.split("\t")
                    chr = lineElems[0]
                    start = lineElems[3]
                    end = lineElems[4] 
                    sumNbelements = lineElems[8].split(";")[2].split("=")[1]
                    #TODO : this is wrong : there is a karyotype input and a data input file , which are different
                    rstart = (float(end)-float(start))/2
                    rend = (float(end)-float(start))/2
                    #outFileHandle.write("%s\t%d\t%d\t%s\n"%(chr,rstart,rend,sumNbelements))
                    outFileHandle.write("%s\t%s\t%s\t%s\n" % (chr, start, end, sumNbelements))
            
#        cmd = "circos -conf etc/circos.conf"
#        os.system(cmd)
        self._log.info("circos plot done")
        
    def _generate_circos_conf(self):
        configFileName = "test_circos.conf"
        with open(configFileName) as cfgFileHandle:
            cfgFileHandle.write("<ideogram>\n")
            cfgFileHandle.write("    show = no\n")
            cfgFileHandle.write("    </ideogram>\n")
            cfgFileHandle.write("chr - domain1 1 0 1000 green\n")
            cfgFileHandle.write("chr - domain2 2 0 1000 red\n")
            

    def _compute_coverage(self, queryFileName, queryFormat, refFileName, refFormat, outputFilename):
        computer = CoverageComputer(self._subscriptVerbosity)
        computer.setInputQueryFile(queryFileName, queryFormat)
        computer.setInputReferenceFile(refFileName, refFormat)
        computer.includeIntrons(False)
        computer.setOutputFileName(outputFilename, title = "TallymerPipe", feature = "Match", featurePart = "Match-Part") 
        computer.run()
        
if __name__ == "__main__":
    iTallymerPipe = TallymerPipe()
    iTallymerPipe.setAttributesFromCmdLine()
    iTallymerPipe.run()