#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""
Connect TE paths via a 'long join' procedure.
"""

import os
import sys
from commons.core.LoggerFactory import LoggerFactory
from commons.core.checker.RepetException import RepetDataException
from commons.core.utils.RepetOptionParser import RepetOptionParser
from commons.core.sql.DbFactory import DbFactory
from commons.core.sql.TablePathAdaptator import TablePathAdaptator
from commons.core.coord.PathUtils import PathUtils
from commons.core.coord.SetUtils import SetUtils
from commons.core.coord.MergedPath import MergedPath

LOG_DEPTH = "repet.annot_pipe"

class LongJoinsForTEs:
    """
    Connect TE paths via a long join procedure.
    """
    
    def __init__( self ):
        """
        Constructor.
        """
        self._inTable = ""
        self._maxOverlap = 15
        self._maxGapLength = 5000
        self._maxMismatchLength = 500
        self._identityTolerance = 2.0
        self._minNestedTEcoverage = 0.95
        self._minLengthToSplit = 100
        self._minLengthToKeepChain = 20
        self._outTable = ""
        self._configFileName = ""
        self._verbose = 0
        self._db = None
        self._tpA = None
        self._tmpTable = ""
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbose)
        self._name = "LongJoinsForTEs"
        
        
    def setAttributesFromCmdLine( self ):
        """
        Set attributes from the command-line arguments.
        """
        usage = "LongJoinForTEs.py [options]"
        description = "Connect TE paths via a long join procedure."
        parser = RepetOptionParser(description = description, epilog = "", usage = usage)
        parser.add_option("-t", "--table",  dest = "inTable",   action = "store",   type = "string",    help = "name of the input table [format='path'] [compulsory]",  default = "")
        parser.add_option("-i", "--identity",   dest = "identityTolerance", action = "store",   type = "float", help = "identity tolerance, in % [default=2]",  default = 2.0)
        parser.add_option("-o", "--overlap",    dest = "maxOverlap",    action = "store",   type = "int",   help = "max overlap size, in bp [default=15]",  default = 15)
        parser.add_option("-g", "--gap",    dest = "maxGapLength",  action="store", type="int", help = "max gap size, in bp [default=5000]",    default = 5000)
        parser.add_option("-m", "--mismatch",   dest = "maxMismatchLength", action="store", type="int", help = "max mismatch size, in bp [default=500]",    default = 500)     
        parser.add_option("-c", "--coverage",   dest = "minNestedTEcoverage",   action="store", type="float",   help = "TE insertion coverage, [default=0.95]", default = 0.95)        
        parser.add_option("-s", "--split",  dest = "minLengthToSplit",  action="store", type="int", help = "min length to split a connection, in bp [default=100]", default = 100)
        parser.add_option("-l", "--keep",   dest = "minLengthToKeepChain",  action="store", type="int", help = "min length to keep a chain [default=20]",   default = 20)
        parser.add_option("-O", "--output", dest = "outTable",  action = "store",   type="string",  help = "name of the output table [default=inTable+'_join']",    default = "")
        parser.add_option("-C", "--config",  dest = "configFileName",   action = "store",   type="string",  help = "configuration file",    default = "")
        parser.add_option("-v", "--verbosity",  dest = "verbosity", action = "store",   type = "int",   help = "verbosity [default: 0, from 0 to 2]",    default = 0)
        options = parser.parse_args()[0]
        self._setAttributesFromOptions(options)
                
    def _setAttributesFromOptions(self, options):
        self.setInputTable(options.inTable)
        self.setIdentityTolerance(options.identityTolerance)
        self.setMaxOverlap(options.maxOverlap)
        self.setMaxGapLength(options.maxGapLength)
        self.setMaxMismatchLength(options.maxMismatchLength)
        self.setMinNestedTEcoverage(options.minNestedTEcoverage)
        self.setMinLengthToSplit(options.minLengthToSplit)
        self.setMinLengthToKeepChain(options.minLengthToKeepChain)
        self.setOutputTable(options.outTable)
        self.setConfigFileName(options.configFileName)
        self.setVerbosityLevel(options.verbosity)
                
    def setInputTable( self, inTable ):
        self._inTable = inTable
        
    def setIdentityTolerance( self, identityTolerance ):
        self._identityTolerance = float(identityTolerance)
        
    def setMaxOverlap( self, maxOverlap ):
        self._maxOverlap = int(maxOverlap)
        
    def setMaxGapLength( self, maxGapLength ):
        self._maxGapLength = int(maxGapLength)
        
    def setMaxMismatchLength( self, maxMismatchLength ):
        self._maxMismatchLength = int(maxMismatchLength)
        
    def setMinNestedTEcoverage( self, minNestedTEcoverage ):
        self._minNestedTEcoverage = float(minNestedTEcoverage)
        
    def setMinLengthToSplit( self, minLengthToSplit ):
        self._minLengthToSplit = int(minLengthToSplit)
        
    def setMinLengthToKeepChain( self, minLengthToKeepChain ):
        self._minLengthToKeepChain = int(minLengthToKeepChain)
        
    def setOutputTable( self, outTable ):
        self._outTable = outTable
        
    def setConfigFileName( self, configFileName ):
        self._configFileName = configFileName
        
    def setVerbosityLevel( self, verbose ):
        self._verbose = int(verbose)
        
    def setLogger( self, logFile="" ):
        if logFile == "":
            logFile = "%s_%s.log" % ( self._inTable, self._name )
        if os.path.exists( logFile ):
            os.remove( logFile )
        self._log = LoggerFactory.createLogger( logFile )
        
        
    def checkAttributes( self ):
        """
        Before running, check the required attributes are properly filled.
        """
        if self._inTable == "":
            self._log.error("missing input table name")
            sys.exit(1)
        if self._outTable == "":
            self._outTable = "%s_join" % self._inTable
            
            
    def getDistanceBetweenRanges( self, r1, r2 ):
        """
        Return the distance between both ranges by handling all possible strands.
        """
        if r1.isOnDirectStrand() != r2.isOnDirectStrand():
            self._log.error("can't compute distance for ranges on different strands")
            sys.exit(1)
        if r1.isOnDirectStrand() and r2.isOnDirectStrand():
            if r1.start < r2.start:
                return r2.start - r1.end
            else:
                return r1.start - r2.end
        else:
            if r1.end < r2.end:
                return r2.end - r1.start
            else:
                return r1.end - r2.start
            
            
    def getDiagonalsGapAndMismatch( self, path1, path2, distQ, distS ):
        """
        Return the diagonals, gap and mismatch between two path instances.
        """
        diag1 = path1.range_subject.getMax() - path1.range_query.getMax()
        diag2 = path2.range_subject.getMin() - path2.range_query.getMin()
        gap = diag1 - diag2
        if gap > 0:  # gap on query
            mismatch = distS
        else:
            mismatch = distQ
        return diag1, diag2, gap, mismatch
    
    
    def joinTEpathsPerQuerySubject( self, qry, sbj, lPaths, lJoinedPaths ):
        """
        Try to join TE paths for a given TE family.
        @param qry: query name
        @type qry: string
        @param sbj: subject name
        @type sbj: string
        @param lPaths: list of paths to join, if possible
        @type lPaths: list of {Path<Path>} instances
        @param lJoinedPaths: list of {Path<Path>} instances that have already been joined
        @type lJoinedPaths: list of {Path<Path>} instances
        """
        if len(lPaths) < 2: return
        
        for p in lPaths:
            if not p.range_query.isOnDirectStrand():
                self._log.error("query in table is on reverse strand")
                sys.exit(1)
                
        for i in xrange(1,len(lPaths)): # try to join path[i] with path[i-1]
            if lPaths[i].range_query.start == lPaths[i-1].range_query.start: # Deny long join: same query start
                continue
            if lPaths[i].range_query.end == lPaths[i-1].range_query.end: # Deny long join: same query end
                continue
            
            distBetweenQueries = self.getDistanceBetweenRanges( lPaths[i-1].range_query, lPaths[i].range_query )
            distBetweenSubjects = self.getDistanceBetweenRanges( lPaths[i-1].range_subject, lPaths[i].range_subject )
            
            if distBetweenQueries > 100000:  # Deny long join: query gap > 100 kb
                continue
            if distBetweenSubjects > 30000:  # Deny long join: subject gap> 30 kb
                continue
            
            if distBetweenQueries < (self._maxOverlap * -1): # Deny long join: too much overlap on query
                continue
            if distBetweenSubjects < (self._maxOverlap * -1): # Deny long join: too much overlap on subject
                continue
            
            diag1, diag2, gap, mismatch = self.getDiagonalsGapAndMismatch( lPaths[i-1], lPaths[i], distBetweenQueries, distBetweenSubjects )

            if mismatch > self._maxMismatchLength: # Deny long join: too much mismatch on query or subject
                continue
            
            try:
                meanIdentity = PathUtils.getIdentityFromPathList( lPaths[ i-1 : i+1 ], True )
            except RepetDataException:
                meanIdentity = 100

            if abs( lPaths[i-1].identity - meanIdentity ) > self._identityTolerance: # Deny long join: break identity tolerance
                continue
            if abs( lPaths[i].identity - meanIdentity ) > self._identityTolerance:
                continue
            
            else:
                if gap <= self._maxGapLength: 
                    # Allow long join: good identity for path[i-1] and path[i] and gp small enough
                    self._tpA.joinTwoPaths( lPaths[i-1].id, lPaths[i].id )
                    lJoinedPaths.append( ( qry, lPaths[i-1], lPaths[i], distBetweenQueries, distBetweenSubjects ) )
                    continue
                
                lNestedPaths = self._tpA.getPathListIncludedInQueryCoord( qry, lPaths[i-1].range_query.end+1, lPaths[i].range_query.start-1 )
                if lNestedPaths != []: # Allow long join: there are nested TEs to justify the gap
                    try:
                        identityNestedPaths = PathUtils.getIdentityFromPathList( lNestedPaths, False )
                    except RepetDataException:
                        identityNestedPaths = 100
                    try:
                        identityCheckedPaths = PathUtils.getIdentityFromPathList( [ lPaths[i-1], lPaths[i] ], True )
                    except RepetDataException:
                        identityCheckedPaths = 100
                    if identityNestedPaths < identityCheckedPaths: # Deny long join: nested TEs are older
                        continue
                    
                    lNestedSets = PathUtils.getSetListFromQueries( lNestedPaths )
                    lMergedNestedSets = SetUtils.mergeSetsInList( lNestedSets )
                    size = SetUtils.getCumulLength( lMergedNestedSets )
                    # TODO: check if it's possible ?
                    if distBetweenQueries <= 0:
                        nestedTEcoverage = 1
                    else :
                        nestedTEcoverage = float(size) / float(distBetweenQueries)
                   
                    if nestedTEcoverage < self._minNestedTEcoverage: # Deny long join: too low coverage of younger, nested TEs
                        continue
                    
                    # Allow long join: nested TEs with high coverage
                    self._tpA.joinTwoPaths( lPaths[i-1].id, lPaths[i].id )
                    lJoinedPaths.append( ( qry, lPaths[i-1], lPaths[i], distBetweenQueries, distBetweenSubjects ) )
                    
                    
    def join( self ):
        """
        Try to join path ranges from a TE annotation table (output from Matcher).
        """
        self._log.info("Beginning of the 'Long Join' procedure...")
        self._log.info("max overlap length: %d" % self._maxOverlap)
        self._log.info("max gap length: %d" % self._maxGapLength)
        self._log.info("max mismatch length: %d" % self._maxMismatchLength)
        self._log.info("identity tolerance: %.2f" % self._identityTolerance)
        self._log.info("inserted TE coverage: %.2f" % self._minNestedTEcoverage)
        
        self._tmpTable = self._tpA.path2PathRange()
        lJoinedPaths = []
        
        lQueryNames = self._tpA.getQueryList()
        lQueryNames.sort()
        for qry in lQueryNames:
            
            lSubjectNames = self._tpA.getSubjectListFromQuery( qry)
            lSubjectNames.sort()
            for sbj in lSubjectNames:
                
                self._tpA._table = self._tmpTable   # get data as 'pathranges'
                lPaths = self._tpA.getPathListWithDirectQueryDirectSubjectFromQuerySubject( qry, sbj )
                self._tpA._table = self._inTable    # do the joins on the input table
                self.joinTEpathsPerQuerySubject( qry, sbj, lPaths, lJoinedPaths )
                
                self._tpA._table = self._tmpTable
                lPaths = self._tpA.getPathListWithDirectQueryReverseSubjectFromQuerySubject( qry, sbj )
                self._tpA._table = self._inTable
                self.joinTEpathsPerQuerySubject( qry, sbj, lPaths, lJoinedPaths )
                
        self._log.info("nb of long joins: %i" % len(lJoinedPaths))
            
            
    def split( self ):
        """
        Try to split the connections made previously by considering all the TE families.
        A connection is cut if an older copy overlaps with a more recent hit (using weighted identity).
        """
        self._log.info("Beginning of the 'Split' procedure...")
        self._log.info("min length to split: %i" % self._minLengthToSplit)
        
        self._db.dropTable(self._outTable)
        self._db.createTable(self._outTable, "path")
        tpAout = TablePathAdaptator(self._db, self._outTable)
        nbSplits = 0
        newId = self._tpA.getNewId()
        
        lQueryNames = self._tpA.getQueryList()
        lQueryNames.sort()
        for qry in lQueryNames:
            lQryPaths = self._tpA.getPathListFromQuery(qry)
            dTmp = dict((iPath.getIdentifier(), MergedPath()) for iPath in lQryPaths)
            for path in lQryPaths:
                dTmp[path.getIdentifier()].add(path)
            lMergedPaths = dTmp.values()
            lMergedPaths.sort(key=lambda x: x.getWeightedIdentity())
            self._log.info("query '%s': %i path collection(s)" % (qry, len(lMergedPaths)))

            for iMergedPath in lMergedPaths:
                lPathFragments = iMergedPath.getPathsList()
                initId = iMergedPath.getID()
                
                lPathsToInsert = []
                if iMergedPath.getNbPaths() == 1: # no connection to test
                    lPathsToInsert += lPathFragments
                    
                else:   # several connected paths, try to split them (cut the connections)
                    lJoinCoordinates = PathUtils.getListOfJoinCoordinatesOnQuery(lPathFragments, self._minLengthToSplit)
                    if lJoinCoordinates == []: # join distance too small to split
                        lPathsToInsert += lPathFragments
                    else:
                        lOverlapPaths = []
                        for joinCoords in lJoinCoordinates:
                            # "Nested paths" might be more descriptive than "overlap" ?
                            lTmp = tpAout.getPathListIncludedInQueryCoord( qry, joinCoords[0], joinCoords[1] )
                            if lTmp != []:
                                if PathUtils.getLengthOnQueryFromPathList( lTmp ) > self._minLengthToSplit:
                                    lOverlapPaths += lTmp

                        if lOverlapPaths == []: # No overlapped path long enough to trigger a split
                            lPathsToInsert += lPathFragments
                        else:
                            lLists = PathUtils.getPathListUnjoinedBasedOnQuery( lOverlapPaths, lPathFragments )
                            nbSplits += len(lLists) - 1
                            lPathsToInsert += lLists[0]
                            for lPaths in lLists[1:]:
                                PathUtils.changeIdInList( lPaths, newId )
                                newId += 1
                                lPathsToInsert += lPaths
                lFilteredPathsToInsert = PathUtils.filterPathListOnChainLength( lPathsToInsert, self._minLengthToKeepChain )
                tpAout.insertList( lFilteredPathsToInsert )
                self._tpA.deleteFromId( initId )

        self._log.info("nb of splits: %i" % nbSplits)
        if not self._tpA.isEmpty():
            self._log.error("remaining paths in temporary table")
            sys.exit(1)
            
            
    def start( self ):
        """
        Open the connection with MySQL, copy the input table and print some stats.
        """
        self.checkAttributes()
        self.setLogger()
        self._log.info("START %s.py" % self._name)
        
        if self._configFileName != "" and os.path.exists(self._configFileName):
            self._db = DbFactory.createInstance(self._configFileName)
        else:
            self._db = DbFactory.createInstance()
        self._db.copyTable(self._inTable, "%s_copy" % self._inTable)
        self._db.changePathQueryCoordinatesToDirectStrand(self._inTable)
        self._tpA = TablePathAdaptator(self._db, self._inTable)
        self._log.info("nb of distinct queries: %i" % len(self._tpA.getQueryList()))
        self._log.info("nb of distinct paths: %i" % self._tpA.getSize())
        self._log.info("nb of distinct identifiers: %i" % self._tpA.getNbIds())

        
    def finish( self ):
        """
        Show some stats, clean temporary tables, close the connection with MySQL.
        """
        self._tpA._table = self._outTable
        self._log.info("nb of distinct queries: %i" % len(self._tpA.getQueryList()))
        self._log.info("nb of distinct paths: %i" % self._tpA.getSize())
        self._log.info("nb of distinct identifiers: %i" % self._tpA.getNbIds())
        self._db.dropTable(self._inTable)
        self._db.dropTable(self._tmpTable)
        self._db.renameTable("%s_copy" % self._inTable, self._inTable)
        self._db.optimizeTable(self._inTable)
        self._db.close()
        
        self._log.info("END %s.py" % self._name)        
        
        
    def run( self ):
        """
        Run the whole program.
        """
        LoggerFactory.setLevel(self._log, self._verbose)
        self.start()
        self.join()
        self.split()
        self.finish()
        
        
if __name__ == "__main__":
    iLongJoinForTEs = LongJoinsForTEs()
    iLongJoinForTEs.setAttributesFromCmdLine()
    iLongJoinForTEs.run()
