#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import os
import linecache
import ConfigParser
from ConfigParser import MissingSectionHeaderError
from commons.core.utils.RepetOptionParser import RepetOptionParser
from commons.core.utils.FileUtils import FileUtils
from commons.core.sql.DbFactory import DbFactory
from commons.core.LoggerFactory import LoggerFactory
from commons.core.coord.AlignUtils import AlignUtils
from commons.core.coord.ConvCoord import ConvCoord
from commons.core.coord.PathUtils import PathUtils
from commons.core.sql.TableSeqAdaptator import TableSeqAdaptator
from commons.tools.getCumulLengthFromTEannot import getCumulLengthFromTEannot
#from SMART.Java.Python.ComputeCoverage import CoverageComputer

LOG_DEPTH = "repet.tools"

## Splices annotations from genome
#
class ComputeCoverageFromAlignAnnotation(object):
    
    def __init__(self, inputFileName = "", configFileName = "", verbose = 3):
        self.ChkMapTableName = ""
        self.projectName = ""
        self.setInputFileName(inputFileName)
        self.setConfigFileName(configFileName)
        self._verbosity = verbose
        self.outputFileName = "" 
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbosity)
        
    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise Exception(errorMsg)
    
    def setAttributesFromCmdLine(self):
        usage = "ComputeCoverageFromAlignAnnotation.py [options]"
        desc = "From align files annotations on chunks, compute the coverage on chromosomes.\n"
        examples = "Example : with the alignAnnotation.align, the coverage.txt output file will contain the total coverage and coverage by each query:\n"
        examples += "\t$ ComputeCoverageFromAlignAnnotation.py -i alignAnnotation.align -C config.cfg -v 1\n"
        examples += "\t Results are in coverage.txt file with coverage calculation\n"
        parser = RepetOptionParser(description = desc, epilog = examples, usage = usage)
        parser.add_option("-i", "--file",       dest = "inputFileName", action = "store", type = "string", help = "input file [mandatory]", default = "")
        parser.add_option("-C", "--config",     dest = "configFileName",     action = "store", type = "string", help = "configuration file name to set database connection [mandatory]", default = "")
        parser.add_option("-v", "--verbose",    dest = "verbose",            action = "store", type = "int",    help = "verbosity level [default: 3]", default = 3)
        options = parser.parse_args()[0]
        self._setAttributesFromOptions(options)
        
    def _setAttributesFromOptions(self, options):
        self.setInputFileName(options.inputFileName)
        self.setConfigFileName(options.configFileName)
        self.setVerbose(options.verbose)
        
    def setInputFileName(self, inputFileName):
        self._inputFileName = inputFileName
            
    def setConfigFileName(self, configFileName):
        self._configFileName = ""
        if configFileName:
            self._configFileName = configFileName
            config = ConfigParser.ConfigParser()

            try :
                config.readfp(open(self._configFileName))
            except MissingSectionHeaderError:
                self._logAndRaise("Configuration file '" + self._configFileName + "' must begin with a section header")
    
            self.setup_env(config)
            self.projectName = config.get("project", "project_name")
            self.ChkMapTableName = "%s_chk_map" % self.projectName
            self.ChrSeqTableName = "%s_chr_seq" % self.projectName
            self.outputFileName = "%s_coverage.txt" % self.projectName
            
    def setVerbose(self, verbose):
        self._verbosity = verbose
        
    def setup_env(self, config):
        os.environ["REPET_HOST"] = config.get("repet_env", "repet_host")
        os.environ["REPET_USER"] = config.get("repet_env", "repet_user")
        os.environ["REPET_PW"] = config.get("repet_env", "repet_pw")
        os.environ["REPET_DB"] = config.get("repet_env", "repet_db")
        os.environ["REPET_PORT"] = config.get("repet_env", "repet_port")

    def checkOptions(self):
        if self._inputFileName:
            if not FileUtils.isRessourceExists(self._inputFileName):
                self._logAndRaise("Input file '%s' does not exist!" % self._inputFileName)
        else:
            self._logAndRaise("No specified mandatory '-i' option!")
       
        if self._configFileName != "":
            self._iDb = DbFactory.createInstance(self._configFileName)
        else:
            self._logAndRaise("No specified configuration file!")

        if not self._iDb.doesTableExist(self.ChkMapTableName):
            self._logAndRaise("Map table on chunk  %s doesn't exist!" % self.ChkMapTableName)

        if not self._iDb.doesTableExist(self.ChrSeqTableName):
            self._logAndRaise("Seq table on chromosome  %s doesn't exist!" % self.ChrSeqTableName)
                    
    def run(self):
        LoggerFactory.setLevel(self._log, self._verbosity)
        self.checkOptions()
        self._log.info("START ComputeCoverageFromAlignAnnotation")
        self._log.info("Convert align file into path file")
        pathFileName = os.path.basename(self._inputFileName)
        pathFileName = pathFileName.replace('.align','.path')
        AlignUtils.convertAlignFileIntoPathFile( self._inputFileName, pathFileName )

        self._log.info("Convert chunk to chromosome")
        pathFileName_onChr = os.path.basename(pathFileName)
        pathFileName_onChr = pathFileName_onChr.replace('.path','_chr.path')
        iConvCoord = ConvCoord(configFile=self._configFileName)
        iConvCoord.setInputData(pathFileName)
        iConvCoord.setInputFormat('path')
        iConvCoord.setMapData(self.ChkMapTableName)
        iConvCoord.setCoordinatesToConvert("q")
        iConvCoord.setMergeChunkOverlaps("yes")
        iConvCoord.setOutputData(pathFileName_onChr)
        iConvCoord.run()

        self._log.info("Create path table on Chr from path file")
        pathTableName_onChr = pathFileName_onChr.replace('.', '_')
        iDB = DbFactory.createInstance(self._configFileName)
        iDB.createTable(pathTableName_onChr, "path", pathFileName_onChr, True)
        iDB.close()
     
        self._log.info("Compute coverage on genome")
        tsA = TableSeqAdaptator( self._iDb, self.ChrSeqTableName )
        genomeLength = tsA.getCumulativSeqLength()
        iGCLFT = getCumulLengthFromTEannot()
        iGCLFT.setInputTable(pathTableName_onChr)
        iGCLFT.setGenomeLength(genomeLength)
        iGCLFT.setConfigFileName(self._configFileName)
#        iGCLFT.setVerbosityLevel(2)
        iGCLFT.setAdaptatorToTable()
        mapFileName = iGCLFT.getAllSubjectsAsMapOfQueries()
        mergeFileName = iGCLFT.mergeRanges( mapFileName )
        totalCoverage = iGCLFT.getCumulLength( mergeFileName )

#        iGCLFT.run()
        coverageFile = open( "coverage.txt", "w" )
        coverageFile.write("Total coverage is %s bp and pct is %.2f %%\n" % (totalCoverage, 100 * totalCoverage / float(genomeLength)))
        dQuery2Coverage = iGCLFT.getdCumulLenthByQuery()
        for key, value in dQuery2Coverage.iteritems():
            coverageFile.write("%s\t%s bp\n" % (key,value))
        coverageFile.close()

        self._log.info("END ComputeCoverageFromAlignAnnotation")
        return 0

if __name__ == '__main__':
    iSFA = ComputeCoverageFromAlignAnnotation()
    iSFA.setAttributesFromCmdLine()
    iSFA.run()
