from commons.core.seq.BioseqDB import BioseqDB
from commons.core.sql.DbFactory import DbFactory
from PASTEC.DecisionRulesParser import DecisionRulesParser
import os

class Environment(object):
    
    def __init__(self, consensusFileName, projectName = "", iConfig = None, decisionRulesFileName = "", verbose = 0):
        #TODO: Bioseq instance : sequence unused, but length needed => create an other class (name + length)
        iBioseqDB = BioseqDB(consensusFileName)
        self._iBioseqDB = iBioseqDB
        self._iDb = None
        self._maxProfilesEvalue = 1e-3
        self._minTeProfilesCoverage = 20
        self._minHgProfilesCoverage = 75
        self._maxHelitronExtremitiesEvalue = 1e-3
        self._minTeBankCoverage = 5
        self._minHgBankCoverage = 95
        self._minRdnaBankCoverage = 95
        self._minHgBankIdentity = 90
        self._minRdnaBankIdentity = 90
        self._minSsrCoverage = 0.75
        self._maxSsrSize = 100
        
        self._blackBoardDemands = []
        self._blackBoardResults = []
                
        self._projectName = projectName
        self._iConfig = iConfig
        self._decisionRulesFileName = decisionRulesFileName
        self._verbosity = verbose
        
        self.setupDecisionRules()
        self.setEnvironementDecisionRules()
        self.setTablesList()
        
    def setTablesList(self):
        #TODO: allow to give a config ! => give a configFileName to Environment, and not a config instance...
        iDb = DbFactory.createInstance()
        self._lTables = iDb.getTableListFromPattern(self._projectName + "%")
        iDb.close()
        
    def setDbInstance(self, iDb):
        self._iDb = iDb
        
    def getBioseqDBInstance(self):
        return self._iBioseqDB
    
    def getTESignifCategoryORFBounds(self):
        return self._TESignifCategoryORFBounds
    
    def getBlastIdWeight(self):
        return self._blastIdWeight
    
    def getProfilesScoreWeight(self):
        return self._profilesScoreWeight
    
    def getMaxProfilesEvalue(self):
        return self._maxProfilesEvalue
    
    def getMinProfilesIndex(self):
        return self._minTeProfilesCoverage
    
    def getMinHGProfilesCoverage(self):
        return self._minHgProfilesCoverage
    
    def getMaxBlastnEvalue(self):
        return self._maxHelitronExtremitiesEvalue
    
    def getMinBlastIndex(self):
        return self._minTeBankCoverage
    
    def getMinHGCoverage(self):
        return self._minHgBankCoverage
    
    def getMinrDNACoverage(self):
        return self._minRdnaBankCoverage
    
    def getMinHGIdentity(self):
        return self._minHgBankIdentity
    
    def getMinrDNAIdentity(self):
        return self._minRdnaBankIdentity
    
    def getMinSSRCoverage(self):
        return self._minSsrCoverage
    
    def getMaxSSRsize(self):
        return self._maxSsrSize
    
    def getMaxCI(self):
        return self._maxCI
    
    def getConsensusNameList(self):
        return self._iBioseqDB.getHeaderList()
    
    def getBlackBoardResults(self):
        return self._blackBoardResults
   
    def getBlackBoardDemands(self):
        return self._blackBoardDemands

    def getDbInstance(self):
        return self._iDb
    
    def getProjectName(self):
        return self._projectName
    
    def getTablesList(self):
        return self._lTables
    
    def getResultsOfOneAgent(self, agentName):
        listOfResultForSpecifiedAgent = []
        for result in self._blackBoardResults:
            if result.demand.agent == agentName:
                listOfResultForSpecifiedAgent.append(result)
        return listOfResultForSpecifiedAgent
    
    def addResultToBlackBoard(self, result):
        self._blackBoardResults.append(result)
   
    def isDemandOnBlackBoardResults(self, demand):
        for res in self._blackBoardResults:
            if res.demand == demand:
                return True
        return False
     
    def isResultOnBlackBoard(self, result):
        return result in self._blackBoardResults
        
    def isConsensusInEnvironment(self, consensusName):
        return consensusName in self.getConsensusNameList()
    
    def displayBlackBoardResult(self):
        for res in self._blackBoardResults:
            if hasattr(res,"getResult"):
                if isinstance(res.getResult(),list):
                    for item in res.getResult():
                        print item
                else:
                    print res.getResult()
            else:
                print(res)
        
    def getResultAccordingToDemand(self, iDemand):
        listOfResults = self.getBlackBoardResults()
        #TODO: find a way NOT to iterate over ALL the list of results : hints : use dictionaries, storing
        #blackboard results by consensus and/or agent type etc
        for result in listOfResults:
            if result.demand == iDemand:
                return result
        raise Exception("The result corresponding at %s-%s is not found on blackboard\n" % (iDemand.agent, iDemand.consensusName))
    
    def getConsensusLength(self, consensusName):
        return self._iBioseqDB.getSeqLength(consensusName)
    
    def _createAttrNameFromConfigKey(self, configKey):
        attributeName = configKey.replace("_", " ")
        attributeName = attributeName.title()
        attributeName = attributeName.replace(" ","") 
        attributeName = "_" + attributeName[0].lower() + attributeName[1:]
        return attributeName
    
    def setAttributesFromConfig(self):
        if self._projectName == "":
            self._projectName = self._iConfig.get("project", "project_name")
        sectionName = "classif_consensus"
        for optionName in self._iConfig.options(sectionName):   
            attributeName = self._createAttrNameFromConfigKey(optionName)
            if hasattr(self, attributeName) and self._iConfig.get(sectionName, optionName) != "":
                try: 
                    setattr(self, attributeName, float(self._iConfig.get(sectionName, optionName)))
                except:   
                    print "Error setting the parameter inside PASTEC's ENVIRONMENT %s" % attributeName
            elif self._verbosity > 3:
                print "The parameter %s does not exist inside PASTEC's ENVIRONMENT" % attributeName

    def setupDecisionRules(self):
        import inspect
        pathToModule = os.path.dirname(inspect.getfile(inspect.currentframe()))
        if self._decisionRulesFileName == "":
            self._decisionRulesFileName = os.path.join(pathToModule, "Decision_rules.yaml")
        self.iDecisionRulesParser = DecisionRulesParser(self._decisionRulesFileName)

    def setEnvironementDecisionRules(self):
        iDecisionRules = self.getAgentDecisionRules("ENVIRONMENT")
        self._TESignifCategoryORFBounds = iDecisionRules["TE_SIGNIF_CATEGORY_ORF_BOUNDS"]
        self._blastIdWeight = iDecisionRules["BLAST_ID_WEIGHT"]
        self._profilesScoreWeight = iDecisionRules["PROFILES_SCORE_WEIGHT"]
        self._maxCI = iDecisionRules["MAX_CI"]

    def getAgentDecisionRules(self, agentName):
        return self.iDecisionRulesParser.getAgentRules(agentName)
        
        