#!/usr/bin/env python

"""
Launch Jellyfish's sub programs, generate map file, and convert output to gff and wig etc
"""

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import re
import os
import time
import shutil
import subprocess
from commons.core.seq.BioseqDB import BioseqDB
from commons.core.seq.BioseqUtils import BioseqUtils
from commons.core.LoggerFactory import LoggerFactory
from commons.core.checker.CheckerUtils import CheckerUtils
from commons.core.utils.RepetOptionParser import RepetOptionParser
from Tallymer_pipe.PlotBenchMarkGFFFiles import PlotBenchMarkGFFFiles
from SMART.Java.Python.convertTranscriptFile import ConvertTranscriptFile

LOG_DEPTH = "repet.tools"

class LaunchJellyfish(object):
    """
    Launch Jellyfish's sub programs, generate map file, and convert output to
    gff and wig etc
    """
    
    _lValidFormats = ["gff", "gff3", "wig", "bed", "map"]
    
    def __init__(self, inputFasta="", indexationFasta=None, merSize=20, minOccs=4, outputFormats="gff", nLargestScaffoldsToPlot=0, threads=1, clean=False, verbosity=0):
        self.inputFasta = inputFasta
        self.indexationFasta = indexationFasta if indexationFasta != None else inputFasta 
        self.merSize = merSize
        self.minOccs = minOccs
        self.outputFormats = outputFormats
        self.nLargestScaffoldsToPlot = nLargestScaffoldsToPlot
        self.doClean = clean
        self.verbosity = verbosity
        
        self.threads = threads
        
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self.verbosity)
        self._workdir = os.path.join(os.getcwd(), "LaunchJellyfish%s" % time.strftime("%Y%m%d%H%M%S"))
        self._tmpSearchFileName = None
        self._tmpMapFileName = None
        self._tmpStatFileName = None   
        self._tmpPngFileName = None
        self._plot_data = {}
        self._plot_data2 = {}
        
    def setAttributesFromCmdLine(self):
        usage = "LaunchJellyfish.py [options]"
        description = "Launch Jellyfish's sub programs, generate map file, and convert output to gff and wig etc, and generates stats from the results.\n"    
        parser = RepetOptionParser(description=description, usage = usage)
        parser.add_option("-i", "--inputFasta", dest="inputFasta", default = "",  action="store", type="string", help="input fasta file [compulsory] [format: fasta]")
        parser.add_option("-u", "--indexationFasta", dest="indexationFasta", default = "",  action="store", type="string", help="input indexation fasta file used to generate kmer index (defaults to input fasta)[optional] [format: fasta]")
        parser.add_option("-s", "--merSize", dest="merSize", default = 20,  action="store", type="int", help="input merSize [optional][default:20]")
        parser.add_option("-m", "--minOccs", dest="minOccs", default = 4,  action="store", type="int", help="input minimal kmer occurence count [default:4]")
        parser.add_option("-f", "--outputFormats", dest="outputFormats", default = "gff",  action="store", type="string", help="comma separated list of output file formats (can  be %s) [optional] [default:gff]" % ", ".join(self._lValidFormats))
        parser.add_option("-n", "--nLargestScaffoldsToPlot",default = 0, type="int", action="store", dest = "nLargestScaffoldsToPlot", help = "generate graph of Kmer repartition along the input sequence for the n biggest scaffolds")
        
        parser.add_option("-t", "--threads", default = 1, type="int", action="store", dest = "threads", help = "number of threads to use")

        parser.add_option("-c", "--clean", dest = "clean", help = "clean temporary files", default = False, action="store_true")
        parser.add_option("-v", "--verbosity", dest="verbosity", default = 1,  action="store", type="int", help="verbosity [optional] ")
        options, args = parser.parse_args()
        self._setAttributesFromOptions(options)
        
    def _setAttributesFromOptions(self, options):
        self.inputFasta = options.inputFasta
        self.indexationFasta = options.indexationFasta if options.indexationFasta != "" else options.inputFasta
        self.merSize = options.merSize
        self.minOccs = options.minOccs
        self.outputFormats = options.outputFormats
        self.nLargestScaffoldsToPlot = options.nLargestScaffoldsToPlot
        self.threads = options.threads
        self.doClean = options.clean    
        self.verbosity = options.verbosity
        
    def _checkOptions(self):
        if self.inputFasta == "":
            self._logAndRaise("Error: missing input fasta file")
        if self.merSize < 1:
            self._logAndRaise("Error: invalid kmer size '%i'; must be a positive integer" % self.merSize)
            
        self.outputFormats = self.outputFormats.lower().split(',')
        sOutFormats = set(self.outputFormats)
        sValidFormats = set(self._lValidFormats)
        lInvalidFormats = list(sOutFormats - sValidFormats)
        self.outputFormats = list(sValidFormats.intersection(sOutFormats))
        if lInvalidFormats:
            self._log.warning("Warning: ignoring invalid formats: <%s>" % " ".join(lInvalidFormats))
        if not self.outputFormats:
            self._logAndRaise("Error: no valid output formats specified")

    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise Exception(errorMsg)
    
    def clean(self):
        try:
            shutil.rmtree(self._workdir)
        except Exception as inst: 
            self._log.error(inst)
                    
    def run(self):
        LoggerFactory.setLevel(self._log, self.verbosity)
        if not CheckerUtils.isExecutableInUserPath("jellyfish"):
            self._logAndRaise("ERROR: 'jellyfish' must be in your path")
        self._checkOptions()
        self._log.debug("Input fasta file: %s; K-mer size: %s; Output formats: %s; Cleaning: %s" % (self.inputFasta, self.merSize, str(self.outputFormats), self.doClean))
        try:
            os.makedirs(self._workdir)
        except:pass
        
        srcPath = os.path.abspath(self.inputFasta)
        dstPath = os.path.join(self._workdir,os.path.basename(self.inputFasta))
        os.symlink(srcPath, dstPath)
        os.chdir(self._workdir)
        
        self.inputFasta = os.path.basename(self.inputFasta)
        self._tmpSearchFileName = "%s.jellyfish" % os.path.splitext(os.path.basename(self.inputFasta))[0]
        self._tmpMapFileName = "%s_tmp.map" % self._tmpSearchFileName
        self._tmpStatFileName = "%s.stat" % self._tmpSearchFileName
        self._tmpPngFileName = "%s.png" % self._tmpSearchFileName
        
        self._runJellyfishCount()
        self._computeOccurences()
        self._convertJellyfishToMap()
        self._writeOutputFiles()
            
        if self.nLargestScaffoldsToPlot > 0:
            self._doPlot()
            shutil.copy2(self._tmpPngFileName, "../.")
        
        shutil.copy2(self._tmpStatFileName, "../.")
        os.chdir("..")
        
        if self.doClean:
            self.clean()

    def _runJellyfishCount(self):
        self._log.info("Starting to run jellyfish search of sequence %s " % (self.inputFasta))
        
        inputFastFile = open(self.inputFasta)
        self.iInputFastaBioseqDB = BioseqDB()
        self.iInputFastaBioseqDB.read(inputFastFile)
        inputFastFile.close()
        
        self.hashSize = self.iInputFastaBioseqDB.getLength()
        
        #if self._readCounting readsCmd= -C 
        
        cmd = "jellyfish count -m %i -s %s -t %i %s" % (self.merSize, self.hashSize, self.threads,  self.indexationFasta)
        process = subprocess.Popen(cmd, shell = True)
        self._log.debug("Running jellyfish count with following params %s" % cmd)
        output = process.communicate()
        self._log.debug("Jellyfish count Output:\n%s" % output[0])
        if process.returncode != 0:
            self._logAndRaise("Error in counter kmers in %s" % self.indexationFasta)
            
        #self._countAndIndexKmersForGivenK()
        self._log.info("Finished running jellyfish scan of sequence %s " % (self.inputFasta))
        
    
    def _computeOccurences(self):
        self._log.info("Starting to dump the number of kmer occurence ")
        os.system("jellyfish dump -c -L %i mer_counts.jf > mer_counts_dumps.txt" % self.minOccs)
#         self._log.info("Starting dumping the number of kmer occurence ")
        #os.system("jellyfish query mer_counts.jf AAAAAAAAAAAAAAAAAAAA > query")
        #os.system("jellyfish histo mer_counts.jf > histo")
        #os.system("jellyfish info mer_counts.jf > infos")

    def _convertJellyfishToMap(self):
        self._log.info("Starting to run jellyfish occurences to map conversion")
        
        #naive implementation 
        def find_all(a_str, sub):
            start = 0
            while True:
                start = a_str.find(sub, start)
                if start == -1: return
                yield start
                start += 1 #+= len(sub)

        
        merCountsFileName = "mer_counts_dumps.txt"
        
        #also write stats in same pass ?
        statFile = open(self._tmpStatFileName, "w")
        statFile.write("kmer\tocc\tfreq\n")
        
        with open(self._tmpMapFileName,'w') as tmpMapFile:
            with open(merCountsFileName) as merCountsFile:
                for line in merCountsFile: #only loads one line at a time
                    kmerData = line.split(" ")
                    kmerSeq = kmerData[0].lower()
                    kmerOccs = int(kmerData[1])
                    if kmerOccs < self.minOccs:
                        #print("occs:",kmerOccs, "is below", self.minOccs,"skipping")
                        continue
                    #print("fetching occurence for each supercontig")
                    for seq in  self.iInputFastaBioseqDB.db:
                        seq.lowCase()
                        querySeq = seq.getSequence()
                        seqPlusOccs = "%s_%d" % (kmerSeq,kmerOccs)
                        for occstart in find_all(querySeq, kmerSeq):
                            occstart += 1
                            outputElements = [seqPlusOccs, str(seq.header), str(occstart) , str(occstart+self.merSize), str(kmerOccs) ]
                            tmpMapFile.write("\t".join(outputElements)+"\n")
                        #print("found in",seq.header,"occurence Starts",occsStarts)
        
        statFile.close()
        #self._writeOccurencesNbAndFrequencyOfKmers(totalNbOcc, dKmer2Occ)      
        self._log.info("sorting map data")
        cmd = """sort -t '\t' -k2,2 -k3,3n %s > sorted.map""" % ( self._tmpMapFileName )
        process = subprocess.Popen(cmd, shell = True)
        output = process.communicate()
        if process.returncode != 0:
            self._logAndRaise("Error in sorting map data")
        shutil.move("sorted.map", self._tmpMapFileName)
        self._log.info("finished sorting map data")  
        self._log.info("Finished jellyfish occurences to map conversion")  
        
    
    def _doPlot(self):
        iBioseqDB = BioseqDB()
        iBioseqDB.load(self.inputFasta)
        largest_seqsDb = iBioseqDB.bestLength(self.nLargestScaffoldsToPlot)
        
        for seq in largest_seqsDb.db:
            headerCleaned = seq.header.replace(" ", "_")
            shortFastaName = "%s_%s" % (os.path.basename(self.inputFasta),headerCleaned)
            data = self._plot_data2[seq.header]
            gffPlotter = PlotBenchMarkGFFFiles(yLabel="Number of Kmer Occurences", maxLength=1, title="Kmer repartition along the input sequence: %s; MerSize: %i" % (shortFastaName, self.merSize) )
            gffPlotter.setOutFileName("%s_%s.png" % (self._tmpSearchFileName, headerCleaned))
            gffPlotter._title = "Mers along the input sequence: %s MerSize: %i" % (shortFastaName, self.merSize) 
            gffPlotter._xLabel = "Coordinates along the input sequence (%s)" % shortFastaName
            gffPlotter._rawData = data
            gffPlotter.run()

    def _writeOccurencesNbAndFrequencyOfKmers(self, totalNbOcc, dKmer2Occ):
        with open(self._tmpStatFileName, "w") as statFile:
            statFile.write("kmer\tocc\tfreq\n")
            for kmer in dKmer2Occ.keys():
                statFile.write("%s\t%i\t%.10f\n" % (kmer, dKmer2Occ[kmer], dKmer2Occ[kmer] / float(totalNbOcc)))

    def _writeOutputFiles(self):
        for format in self.outputFormats:
            self._log.info("Generating %s file" % format)
            outputFileName = "%s.jellyfish.%s" % (os.path.splitext(self.inputFasta)[0], format)
            try:
                iConvertTranscriptFile = ConvertTranscriptFile(inputFileName=self._tmpMapFileName, name="Jellyfish",\
                         inputFormat="map", outputFileName=outputFileName, outputFormat=format,feature= "Match", featurePart="Match-part", verbosity=0) #self.verbosity
                iConvertTranscriptFile.run()
            except Exception as inst:
                self._log.error("Error: %s - Failed to generate %s format ouput, skipping" % (inst, format))
            shutil.copy2(outputFileName, "../.")


class ConvertUtils(object):
    
    def convertJellyfishFormatIntoMapFormatAndGenerateData(fastaFileName, searchFileName, mapFileName):
        dIndex2NameLengthList = ConvertUtils._createDictOfNameLengthAccordingToSeqOrder(fastaFileName)
        plotData = {}
        plotData2 = {}
        with open(searchFileName, "r") as talFile:
            with open(mapFileName, "w") as mapFile:
                totalNbOcc = 0
                dKmer2Occ = {}
                line = talFile.readline()
                while line:
                    data = line[:-1].split("\t")
                    name = "%s_%s" % (data[3], data[2])
                    nbOccs = int(data[2])
                    chrname = dIndex2NameLengthList[int(data[0])][0]
                    if data[1][0] == "+":
                        start = int(data[1][1:]) + 1
                        end = start + len(data[3])
                    elif data[1][0] == "-":
                        start_revcomp = int(data[1][1:])
                        start = dIndex2NameLengthList[int(data[0])][1] - start_revcomp - 1
                        end = end - len(data[3]) + 1
                    mapLine = "%s\t%s\t%s\t%s\t%i\n" % (name, chrname, start, end, nbOccs)
                    mapFile.write(mapLine)
                    
                    if plotData2.get(chrname,None) is None:
                        plotData2[chrname] = {}
                    if plotData2[chrname].get(start, None) is None:
                        plotData2[chrname][start]=0
                    plotData2[chrname][start] += nbOccs
                    
                    totalNbOcc += 1
                    if dKmer2Occ.has_key(data[3]):
                        dKmer2Occ[data[3]] += 1
                    else:
                        dKmer2Occ[data[3]] = 1
                    plotData[start] = nbOccs
                    line = talFile.readline()
        return totalNbOcc, dKmer2Occ, plotData, plotData2
    
    convertJellyfishFormatIntoMapFormatAndGenerateData = staticmethod(convertJellyfishFormatIntoMapFormatAndGenerateData)
    
    def _createDictOfNameLengthAccordingToSeqOrder(fastaFileName):
        with open(fastaFileName) as fastaFile:
            line = fastaFile.readline()
            i = 0
            length = 0
            dIndex2Name = {}
            while line:
                if line[0] == ">":
                    dIndex2Name[i] = [line[1:-1]]
                    if i > 0:
                        dIndex2Name[i - 1].append(length)
                        length = 0
                    i += 1
                else:
                    length += len(line[:-1])
                line = fastaFile.readline()
            dIndex2Name[i - 1].append(length)
        return dIndex2Name
    
    _createDictOfNameLengthAccordingToSeqOrder = staticmethod(_createDictOfNameLengthAccordingToSeqOrder)

if __name__ == "__main__":
    iLaunchJellyfish = LaunchJellyfish()
    iLaunchJellyfish.setAttributesFromCmdLine()
    iLaunchJellyfish.run()        