#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import os
import subprocess
from commons.core.LoggerFactory           import LoggerFactory
from commons.core.checker.CheckerUtils    import CheckerUtils
from commons.core.checker.RepetException  import RepetException
from commons.core.utils.FileUtils         import FileUtils
from commons.core.utils.RepetOptionParser import RepetOptionParser

LOG_DEPTH = "commons.launcher"

class LaunchQuast(object):
    
    def __init__(self, assemblyFileName = "", referenceFileName = "", minContigSize = 100, nbThreads = 1, verbosity = 3):
        self._assemblyFileName = assemblyFileName
        self._referenceFileName = referenceFileName
        self._minContigSize = minContigSize
        self._nbThreads = nbThreads
        
        self._verbosity = verbosity
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbosity)
        
    def setAttributesFromCmdLine(self):
        usage = "LaunchQuast.py [options]"
        description = "Launch QUAST: Quality Assessment Tool for Genome Assemblies.\n"
        epilog = "Example 1: launch with guided metrics, with informative verbosity.\n"
        epilog += "\t$ LaunchQuast.py -a assembly.fa -r reference.fa -v 3\n"
        parser = RepetOptionParser(description = description, epilog = epilog, usage = usage)
        parser.add_option("-a", "--assembly",       dest = "assemblyFileName",  action = "store",       type = "string", help = "input assembly file name [compulsory] [format: fasta]", default = "")
        parser.add_option("-r", "--reference",      dest = "referenceFileName", action = "store",       type = "string", help = "input reference file name [compulsory] [format: fasta]", default = "")
        parser.add_option("-m", "--minContigSize",  dest = "minContigSize",     action = "store",       type = "int",    help = "Lower threshold for a contig length [optional] [default:100]", default = 100)
        parser.add_option("-t", "--threads",        dest = "nbThreads",         action = "store",       type = "int",    help = "number of threads to use for aligning [optional] [default:1]", default = 1)
        parser.add_option("-v", "--verbosity",      dest = "verbosity",         action = "store",       type = "int",    help = "verbosity [optional] [default: 3]", default = 3)
        options = parser.parse_args()[0]
        self._setAttributesFromOptions(options)
        
    def _setAttributesFromOptions(self, options):
        self._assemblyFileName = options.assemblyFileName
        self._referenceFileName = options.referenceFileName
        self._minContigSize = options.minContigSize
        self._nbThreads = options.nbThreads
        self.setVerbosity(options.verbosity)
        
    def setVerbosity(self, verbosity):
        self._verbosity = verbosity
        
    def _checkOptions(self):
        if self._assemblyFileName == "":
            self._logAndRaise("ERROR: Missing input assembly file name")
        else:
            if FileUtils.isRessourceExists(self._assemblyFileName):
                self._assemblyFileName = os.path.abspath(self._assemblyFileName)
            else:
                self._logAndRaise("Input assembly file '%s' does not exist!" % self._assemblyFileName)
                
        if self._referenceFileName == "":
            self._log.warning("WARNING: no input assembly file name specified")
        else:
            if FileUtils.isRessourceExists(self._referenceFileName):
                self._referenceFileName = os.path.abspath(self._referenceFileName)
            else:
                self._logAndRaise("Input reference file '%s' does not exist!" % self._referenceFileName)
        
    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise RepetException(errorMsg)
    
    def _parseReport(self):
        stats = {}
        with open("quast_results/latest/report.txt") as quastResultFile:
            for line in  quastResultFile.readlines():
                if "N50" in line:
                    line = line.rstrip()
                    line = line.split(" ")
                    N50 = int(  line.pop() )
                    stats["N50"] = N50
                if "NG50" in line:
                    line = line.rstrip()
                    line = line.split(" ")
                    NG50= int( line.pop() )
                    stats["NG50"] = NG50
                if "Genome fraction" in line:
                    line = line.rstrip()
                    line = line.split(" ")
                    coverage= float( line.pop() )
                    stats["coverage"] = coverage
        return stats
                 
    def run(self):
        LoggerFactory.setLevel(self._log, self._verbosity)
        self._log.info("START LaunchQuast")
        if not CheckerUtils.isExecutableInUserPath("quast.py"):
            self._logAndRaise("ERROR: 'quast.py' must be in your path")
        self._checkOptions()
        
        self._log.debug("Assembly file name: %s" % self._assemblyFileName)
        self._log.debug("Reference file name: %s" % self._referenceFileName)
        
        cmd = "quast.py "
        if self._referenceFileName: cmd += "-R %s " % self._referenceFileName
        cmd += "--min-contig %d " % self._minContigSize
        cmd += "--threads %d " % self._nbThreads
        cmd += "%s" % self._assemblyFileName

        process = subprocess.Popen(cmd, shell = True)
        self._log.debug("Running : '%s'" % cmd)
        process.communicate()
        if process.returncode != 0:
            self._logAndRaise("ERROR when launching '%s'" % cmd)
            
        self._log.debug("Parsing QUAST report")
        stats = self._parseReport()
        
        self._log.info("END LaunchQuast")
        
        return stats

if __name__ == "__main__":
    iLaunchQuast = LaunchQuast()
    iLaunchQuast.setAttributesFromCmdLine()
    iLaunchQuast.run()        