#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import os
import time
import shutil
import subprocess
from commons.core.seq.BioseqDB import BioseqDB
from commons.core.LoggerFactory import LoggerFactory
from commons.core.checker.CheckerUtils import CheckerUtils
from commons.core.utils.RepetOptionParser import RepetOptionParser

LOG_DEPTH = "commons.launcher"

class LaunchRepeatScout(object):
    """Launches RepeatScout program"""
    
    def __init__(self, fastaFileName = "", outFileName = "", minLength = 50, doClean = False, verbosity = 0):
        self._fastaFileName = fastaFileName
        self.setOutFileName(outFileName)
#        self._configFileName = configFileName
        self._minLength = minLength
        self._doClean = doClean
        self._verbosity = verbosity
        self._workdir = os.path.join(os.getcwd(), "LaunchRepeatScout_%s" % time.strftime("%Y%m%d%H%M%S"))
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbosity)
        
    def setAttributesFromCmdLine(self):
        usage = "LaunchRepeatScout.py [options]"
        description = "Launch Repeat Scout program.\n"
        epilog = "\nExample 1: launch without verbosity and keep temporary files.\n"
        epilog += "\t$ LaunchRepeatScout.py -i file.fa -v 0"
        epilog += "\n\t"
        epilog += "\nExample 2: launch with verbosity to have errors (level 1) and basic information (level 2), and delete temporary files.\n"
        epilog += "\t$ LaunchRepeatScout.py -i file.fa -c -v 2\n"
        parser = RepetOptionParser(description = description, epilog = epilog, usage = usage)
        parser.add_option("-i", "--fasta",      dest = "fastaFileName", action = "store",       type = "string", help = "input fasta file name [compulsory] [format: fasta]", default = "")
        parser.add_option("-l", "--minLength",  dest = "minLength",     action = "store",       type = "int",    help = "minimum consensus length [optional] [default: 50]", default = 50)
        
#        parser.add_option("-L", "--extend", dest="extend", default = 10000,  action="store", type="int", help="size of region to extend left or right [optional][default:10000]")
#        parser.add_option("", "--match", dest="match", default = 1,  action="store", type="int", help="reward for a match  [optional][default:+1]")
#        parser.add_option("", "--mismatch", dest="mismatch", default = -1,  action="store", type="int", help=" penalty for a mismatch  [optional][default:-1]")
#        parser.add_option("", "--gap", dest="gap", default = -5,  action="store", type="int", help=" penalty for a gap  [optional][default:-5]")
#        parser.add_option("-g", "--maxgap", dest="maxgap", default = 5,  action="store", type="int", help="maximum number of gaps allowed  [default:5]")
#        parser.add_option("-M", "--maxOccs", dest="maxOccs", default = 10000,  action="store", type="int", help="cap on the number of sequences to align [optional][default:10000]")
#        parser.add_option("-r", "--maxreps", dest="maxreps", default = 10000,  action="store", type="int", help="stop work after reporting this number of repeats  [optional][default:10000]")
#        parser.add_option("", "--capPenalty", dest="capPenalty", default = -20,  action="store", type="int", help="cap on penalty for exiting alignment of a sequence [optional][default:-20]")
#        parser.add_option("-t", "--tandemDist", dest="tandemDistance", default = 500,  action="store", type="int", help="number of bases that must intervene between two l-mers for both to be counted [optional][default:500]")
#        parser.add_option("-m", "--minOccs", dest="minOccs", default = 3,  action="store", type="int", help="stop if fewer than this number of l-mers are found in the seeding phase  [optional][default:3]")
#        parser.add_option("", "--minImp", dest="minImprovement", default = 3,  action="store", type="int", help="mount that a the alignment needs to improve each step to be considered progress  [optional][default:3]")
#TODO give access to this option and the default value at 500
#        parser.add_option("", "--stopafter", dest="stopAfter", default = 100,  action="store", type="int", help="stop the alignment after this number of no-progress columns  [optional][default:100]")
#        parser.add_option("", "--maxEntropy", dest="maxEntropy", default = -0.7,  action="store", type="float", help="entropy (complexity) threshold for an l-mer to be considered [optional][default:-0.7]")
#      
        parser.add_option("-o", "--out",        dest = "outFileName",   action = "store",       type = "string", help = "output file name [default: <input>_RepeatScoutConsensus.fa]", default = "")
        #parser.add_option("-C", "--config",     dest = "configFileName",action = "store",       type = "string", help = "configuration file name (e.g. TEdenovo.cfg)", default = "")
        parser.add_option("-c", "--clean",      dest = "doClean",       action = "store_true",                   help = "clean temporary files [optional] [default: False]", default = False)
        parser.add_option("-v", "--verbosity",  dest = "verbosity",     action = "store",       type = "int",    help = "verbosity [optional] [default: 1]", default = 1)
        options = parser.parse_args()[0]
        self._setAttributesFromOptions(options)
        
    def _setAttributesFromOptions(self, options):
        self.setFastaFileName(options.fastaFileName)
        self.setOutFileName(options.outFileName)
        self.setMinLength(options.minLength)
        #self.setConfigFileName(options.configFileName)
        self.setDoClean(options.doClean)
        self.setVerbosity(options.verbosity)

#    def _checkConfig(self):       
#        iConfigRules = ConfigRules()
#        iConfigRules.addRuleSection(section="", mandatory=True)
#        iConfigRules.addRuleOption(section="", option ="fasta_name", mandatory=True, type="string")
#        iConfigRules.addRuleOption(section="", option ="clean", mandatory=True, type="bool")
#        iConfigChecker = ConfigChecker(self._configFileName, iConfigRules)
#        iConfig = iConfigChecker.getConfig()
#        self._setAttributesFromConfig(iConfig)
#        
#    def _setAttributesFromConfig(self, iConfig):
#        self.setFastaFileName(iConfig.get("", "fasta_name"))
#        self.setOutFileName(self._outFileName)
#        self.setDoClean(iConfig.get("", "clean"))
        
    def setFastaFileName(self, fastaFileName):
        self._fastaFileName = fastaFileName
        
    def setOutFileName(self, outFileName):
        if outFileName == "":
            self._outFileName = "%s_RepeatScoutConsensus.fa" % os.path.splitext(os.path.basename(self._fastaFileName))[0]
        else:
            self._outFileName = outFileName
            
        self._freqFileName = "%s.freq" % os.path.basename(self._fastaFileName)
        self._outputRepeatsFileName = "%s.reps" % os.path.basename(self._fastaFileName)
        self._filteredRepeatsFileName = "%s.reps.filtered" % os.path.basename(self._fastaFileName)
        self._filteredMaskedRepeatsFileName = "%s.out" % os.path.basename(self._fastaFileName)
        self._outputGffFileName = "%s.out.gff3" % os.path.basename(self._fastaFileName)
        
    def setMinLength(self, minLength):
        self._minLength = minLength
        
#    def setConfigFileName(self, configFileName):
#        self._configFileName = configFileName
        
    def setDoClean(self, doClean):
        self._doClean = doClean
        
    def setVerbosity(self, verbosity):
        self._verbosity = verbosity
        
    def _checkOptions(self):
        if self._fastaFileName == "":
            self._logAndRaise("ERROR: Missing input fasta file name")
            
        if self._minLength < 0:
            self._logAndRaise("ERROR: minLength option (-l) must be a strictly positive integer")
            
    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise Exception(errorMsg)
    
    def _runBuildLmerTable(self):
        self._log.info("STARTED build_lmer_table")
        
        cmd = "build_lmer_table"
        cmd += " -sequence %s" % self._fastaFileName
        cmd += " -freq %s" % self._freqFileName
        process = subprocess.Popen(cmd.split(' '), stdout = subprocess.PIPE, stderr = subprocess.PIPE)
        self._log.debug("Running : %s" % cmd)
        output = process.communicate()
        self._log.debug("Output:\n%s" % output[0])
        if process.returncode != 0:
            self._logAndRaise("ERROR when launching '%s'" % cmd)
            
        self._log.info("FINISHED build_lmer_table")
    
    def _runRepeatScout(self):
        self._log.info("STARTED RepeatScout")
        
        cmd = "RepeatScout"
        cmd += " -sequence %s" % self._fastaFileName
        cmd += " -output %s" % self._outputRepeatsFileName
        cmd += " -freq %s" % self._freqFileName
        cmd += " -goodlength %i" % self._minLength
        cmd += ' -'+'v'*(max(self._verbosity-1, 0))
        process = subprocess.Popen(cmd, shell = True)
        self._log.debug("Running : %s" % cmd)
        output = process.communicate()
        self._log.debug("Output:\n%s" % output[0])
        if process.returncode != 0:
            self._logAndRaise("ERROR when launching '%s'" % cmd)
            
        self._log.info("FINISHED RepeatScout")
    
    def _runFilterStage1(self):
        self._log.info("STARTED FilterStage1")
        
        cmd = "filter-stage-1.prl %s > %s" % (self._outputRepeatsFileName, self._filteredRepeatsFileName)
        os.system(cmd)
        self._log.debug("Running : %s" % cmd)
            
        self._log.info("FINISHED FilterStage1")
    
#    def _runRepeatMasker(self):
#        iLaunchRepeatMasker = LaunchRepeatMasker(self._fastaFileName, self._filteredRepeatsFileName,verbosity=1 )
#        iLaunchRepeatMasker.run()    
#    
#    def _convertToGff3(self):
#        self._log.info("STARTED convert to gff3")
#        rmPath = os.environ["REPEATMASKER"] 
#        cmdPath = os.path.join(rmPath,"util","rmOutToGFF3.pl")
#        cmd = "%s %s > %s" % (cmdPath, self._filteredMaskedRepeatsFileName, self._outputGffFileName)
#        os.system(cmd)
#        self._log.debug("Running : %s" % cmd)
#        self._log.info("FINISHED convert to gff3")
#        
#    def _runFilterStage2(self):
#        self._log.info("STARTED FilterStage2")
#        
#        cmd = "filter-stage-2.prl %s > %s" % (self._outputRepeatsFileName, self._filteredRepeatsFileName)
#        os.system(cmd)
#        self._log.debug("Running : %s" % cmd)
#        self._log.info("FINISHED FilterStage2")
    
    def _formatOutputForTEdenovo(self):
        self._log.info("STARTED Fasta formatting")
        projectName = (os.path.basename(self._fastaFileName).split('.')[0]).replace(" ","_")
        
        iBSDB = BioseqDB(self._filteredRepeatsFileName)
        iBSDB.upCase()
        iBSDB.sortByLength(reverse = True)
        
        countSeq = 1
        for iBS in iBSDB.db:
            iBS.setHeader("%s_RS_%d" % (projectName, countSeq))
            countSeq += 1
            
        with open(self._outFileName, "w") as fH:
            iBSDB.write(fH)
        
        shutil.move(self._outFileName, "..")
        os.symlink("../%s" % self._outFileName, os.path.basename(self._outFileName))
        self._log.info("FINISHED Fasta formatting")    
                  
    def run(self):
        LoggerFactory.setLevel(self._log, self._verbosity)
#        if self._configFileName:
#            self._checkConfig()
        if not CheckerUtils.isExecutableInUserPath("RepeatScout"):
            self._logAndRaise("ERROR: 'RepeatScout' must be in your path")
        self._checkOptions()
        self._log.info("START LaunchRepeatScout")
        self._log.debug("Input Fasta file name: %s" % self._fastaFileName)
        
        try:
            os.makedirs(self._workdir)
        except:pass
        
        inFaFilePath = os.path.abspath(self._fastaFileName)
        self._fastaFileName = os.path.basename(inFaFilePath)
        os.chdir(self._workdir)
        os.symlink(inFaFilePath, self._fastaFileName)
        
        self._runBuildLmerTable()
        self._runRepeatScout()
        self._runFilterStage1()
        self._formatOutputForTEdenovo()
        #self._runRepeatMasker()
        #self._convertToGff3()
        
        os.chdir("..")
        if self._doClean:
            self._log.warning("Files will be cleaned")
            os.rmdir(self._workdir)
            
        self._log.info("END LaunchRepeatScout")

if __name__ == "__main__":
    iLaunchRepeatScout = LaunchRepeatScout()
    iLaunchRepeatScout.setAttributesFromCmdLine()
    iLaunchRepeatScout.run()        