#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import os
import sys
import time
import shutil
import subprocess

if not "REPET_PATH" in os.environ.keys():
    print "ERROR: no environment variable REPET_PATH"
    sys.exit(1)
sys.path.append(os.environ["REPET_PATH"])
if not "PYTHONPATH" in os.environ.keys():
    os.environ["PYTHONPATH"] = os.environ["REPET_PATH"]
else:
    os.environ["PYTHONPATH"] = "%s:%s" % (os.environ["REPET_PATH"], os.environ["PYTHONPATH"])

from commons.core.utils.RepetOptionParser import RepetOptionParser
from commons.core.utils.FileUtils import FileUtils
from commons.core.LoggerFactory import LoggerFactory
from commons.core.checker.CheckerUtils import CheckerUtils
from commons.core.checker.ConfigChecker import ConfigRules, ConfigChecker
from commons.core.sql.DbFactory import DbFactory
from commons.core.sql.TablePathAdaptator import TablePathAdaptator
from commons.core.coord.SetUtils import SetUtils
from commons.core.coord.ConvCoord import ConvCoord
from commons.tools.GFF3Maker import GFF3Maker
from commons.tools.PrepareBatches import PrepareBatches
from SegDup_pipe.SelfAlign import SelfAlign

LOG_DEPTH = "repet.SegDup"

class SegDupPipeline(object):
    
    def __init__(self, configFileName = "SegDup.cfg", step = 1, verbosity = 3):
        self._pipelineName = "SegDup"
        self._configFileName = configFileName
        self._step = step
        self._verbosity = verbosity
        self._iConfig = None
        self._projectName = ""
        self._projectDir = ""
        self._inFileName = ""
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbosity)
  
    def setAttributesFromCmdLine(self):
        description = "Launch Segmental Duplications detection pipeline."
        epilog= "\n"
        parser = RepetOptionParser(description = description, epilog = epilog)
        parser.add_option("-C", "--config",    dest = "configFileName", action = "store",       type = "string", help = "input config file name [optional] [format: cfg]",   default = "SegDup.cfg")
        parser.add_option("-S", "--step",      dest = "step",           action="store",         type = "int",    help = "step (1/2/3/4/5)",                                  default = 1)
        parser.add_option("-v", "--verbosity", dest = "verbosity",      action = "store",       type = "int",    help = "verbosity [optional] [default: 3]",                 default = 3)
        options = parser.parse_args()[0]
        self._setAttributesFromOptions(options)
        
    def _setAttributesFromOptions(self, options):
        self._configFileName = options.configFileName
        self._inFileName = "%s.fa" % self._projectName 
        self._step = options.step
        self._verbosity = options.verbosity
        
    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise Exception(errorMsg)
    
    def _checkOptions(self):
        if self._step not in [1,2,3,4,5]:
            self._logAndRaise("ERROR: Bad step '%i'. Must be 1,2,3,4 or 5" % self._step)
            
    def _checkConfig(self):       
        iConfigRules = ConfigRules()
        iConfigRules.addRuleSection(section="repet_env", mandatory=True)
        changelogVersion = CheckerUtils.getVersionFromChangelogFile("%s/doc/CHANGELOG" % os.environ["REPET_PATH"])
        iConfigRules.addRuleOption(section="repet_env", option ="repet_version", set =(changelogVersion), mandatory=True)
        iConfigRules.addRuleOption(section="repet_env", option ="repet_host", mandatory=True, type="string")
        iConfigRules.addRuleOption(section="repet_env", option ="repet_user", mandatory=True, type="string")
        iConfigRules.addRuleOption(section="repet_env", option ="repet_pw", mandatory=True, type="string")
        iConfigRules.addRuleOption(section="repet_env", option ="repet_db", mandatory=True, type="string")
        iConfigRules.addRuleOption(section="repet_env", option ="repet_port", mandatory=True, type="string")
        iConfigRules.addRuleOption(section="repet_env", option ="repet_job_manager", mandatory=True, type="string")
        
        iConfigRules.addRuleSection(section="project", mandatory=True)
        iConfigRules.addRuleOption(section="project", option ="project_name", mandatory=True, type="string")
        iConfigRules.addRuleOption(section="project", option ="project_dir", mandatory=True, type="string")

        iConfigChecker = ConfigChecker(self._configFileName, iConfigRules)
        self._iConfig = iConfigChecker.getConfig()
        self._setAttributesFromConfig()
        
    def _setAttributesFromConfig(self):
        self._projectName = self._iConfig.get("project", "project_name")
        self._projectDir = self._iConfig.get("project", "project_dir")
        
        if not CheckerUtils.isMax15Char(self._projectName):
            self._logAndRaise("ERROR: project name must have 15 character max")
        if not CheckerUtils.isCharAlphanumOrUnderscore(self._projectName):
            self._logAndRaise("ERROR: project name must contain only alphanumeric or underscore character")
          
    def _setupEnv(self):
        os.environ["REPET_HOST"] = self._iConfig.get("repet_env", "repet_host")
        os.environ["REPET_USER"] = self._iConfig.get("repet_env", "repet_user")
        os.environ["REPET_PW"] = self._iConfig.get("repet_env", "repet_pw")
        os.environ["REPET_DB"] = self._iConfig.get("repet_env", "repet_db")
        os.environ["REPET_PORT"] = self._iConfig.get("repet_env", "repet_port")
        os.environ["REPET_JOB_MANAGER"] = self._iConfig.get("repet_env", "repet_job_manager")
        os.environ["REPET_QUEUE"] = self._iConfig.get("repet_env", "repet_job_manager")
        os.environ["REPET_JOBS"] = "MySQL"
        
    def run(self):
        LoggerFactory.setLevel(self._log, self._verbosity)
        self._log.info("START %s (%s)\n" % (self._pipelineName, time.strftime("%Y-%m-%d %H:%M:%S")))
#        self._log.always("START %s (%s)\n" % (self._pipelineName, time.strftime("%Y-%m-%d %H:%M:%S")))

        self._checkOptions()
        self._checkConfig()
        self._setupEnv()
        
        if self._step == 1:
            iPrepareBatches = PrepareBatches(self._pipelineName, self._projectDir, self._projectName, self._iConfig, self._verbosity)
            iPrepareBatches.run()
        elif self._step == 2:
            iSelfAlignAndConvChunkToChr = SelfAlignAndConvChunkToChr(self._pipelineName, self._configFileName, self._verbosity)
            iSelfAlignAndConvChunkToChr.run()
        elif self._step == 3:
            iFilterKnownRepeatsAndCrossMatches = FilterKnownRepeatsAndCrossMatches(self._configFileName, self._verbosity)
            iFilterKnownRepeatsAndCrossMatches.run()
        elif self._step == 4:
            iMatcherAndFilterKnownRepeats = MatcherAndFilterKnownRepeats(self._configFileName, self._verbosity)
            iMatcherAndFilterKnownRepeats.run()
        elif self._step == 5:
            iClusterizeAndRegroupByClusterSize = ClusterizeAndRegroupByClusterSize(self._configFileName, self._verbosity)
            iClusterizeAndRegroupByClusterSize.run()
            
        self._log.info("END %s (%s)" % (self._pipelineName, time.strftime("%Y-%m-%d %H:%M:%S")))
            
class SegDupStep(object):
    
    def __init__(self, configFileName, verbosity = 1):
        self._configFileName = configFileName
        self._verbosity = verbosity
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbosity)

    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise Exception(errorMsg)
    
class SelfAlignAndConvChunkToChr(SegDupStep):
    
    def __init__(self, pipelineName, configFileName, verbosity = 1):
        SegDupStep.__init__(self, configFileName, verbosity)
        self._pipelineName = pipelineName
    
    def _setAttributesFromConfig(self):
        iConfigRules = ConfigRules()
        iConfigChecker = ConfigChecker(self._configFileName, iConfigRules)
        iConfig = iConfigChecker.getConfig()
        self._projectName = iConfig.get("project", "project_name")
        self._projectDir = iConfig.get("project", "project_dir")
        self._doClean = True if iConfig.get("self_align", "clean") == "yes" else False

    def run(self):
        self._log.info("Beginning of step 2")
        self._setAttributesFromConfig()
        
        iSelfAlign = SelfAlign(self._pipelineName, self._configFileName, self._verbosity)
        iSelfAlign.run()
        
        os.chdir("%s_Blaster" % self._projectName)
        shutil.move("%s.align.not_over.filtered" % self._projectName, "%s_chk.align.not_over.filtered" % self._projectName)
        try:
            os.symlink("../%s" % self._configFileName, "%s" % self._configFileName)
        except:
            pass
        
        iConvCoord = ConvCoord()
        iConvCoord.setConfigFile(self._configFileName)
        iConvCoord.setInputData("%s_chk.align.not_over.filtered" % self._projectName)
        iConvCoord.setMapData("%s/%s_db/%s_chunks.map" % (self._projectDir, self._projectName, self._projectName))
        iConvCoord.setCoordinatesToConvert("qs")
        iConvCoord.setMergeChunkOverlaps(False)
        iConvCoord.setVerbosityLevel(self._verbosity - 2)
        iConvCoord.setOutputData("%s_chr.align.not_over.filtered" % self._projectName)
        iConvCoord.run()
        
        if self._doClean:
            os.remove(self._configFileName)
        
        os.chdir(self._projectDir)
        self._log.info("Step 2 finished successfully\n")

class FilterKnownRepeatsAndCrossMatches(SegDupStep):
    
    def __init__(self, configFileName, verbosity = 1):
        SegDupStep.__init__(self, configFileName, verbosity)
        self._sectionName = "filtering"
    
    def _checkConfig(self):       
        iConfigRules = ConfigRules()
        iConfigRules.addRuleSection(section=self._sectionName, mandatory=True)
        iConfigRules.addRuleOption(section=self._sectionName, option ="extend", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._sectionName, option ="TE_annotation_table_names", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._sectionName, option ="Satellites_annotation_table_names", mandatory=True, type="string")

        iConfigChecker = ConfigChecker(self._configFileName, iConfigRules)
        self._iConfig = iConfigChecker.getConfig()
        self._setAttributesFromConfig()
        
    def _setAttributesFromConfig(self):
        self._projectName = self._iConfig.get("project", "project_name")
        self._projectDir = self._iConfig.get("project", "project_dir")
        self._extend = self._iConfig.get(self._sectionName, "extend")
        self._TETableNames = self._iConfig.get(self._sectionName, "TE_annotation_table_names")
        self._SatTableNames = self._iConfig.get(self._sectionName, "Satellites_annotation_table_names")
        
    def _checkAttributes(self):
        if not self._TETableNames and not self._SatTableNames:
            self._logAndRaise("ERROR: you need to specify at least one annotation table in your configuration file")
        
    def run(self):
        self._log.info("Beginning of step 3")
        self._checkConfig()
        self._checkAttributes()
        stepdir = "%s_Blaster_NoRep" % self._projectName
        os.makedirs(stepdir)
        os.chdir(stepdir)
         
        alignFileName = "%s_chr.align.not_over.filtered" % self._projectName
        alignFilePath = "%s/%s_Blaster/%s" % (self._projectDir, self._projectName, alignFileName)
        annotTETableNameNoRepet = "%s_NoRep_%s_align" % (alignFileName.replace('.','_'), self._extend)
        os.symlink(alignFilePath, alignFileName)
        
        self._log.info("START repeats filtering")
        
        cmd = "sqlEliminRepeats.py "
        cmd += "-b %s " % alignFileName
        cmd += "-o %s " % self._extend
        if self._TETableNames != "":
            cmd += "-t "
            cmd += " -t ".join(self._TETableNames.split(","))
        if self._SatTableNames != "":
            cmd += " -s "
            cmd += " -s ".join(self._SatTableNames.split(","))
        cmd += " -e %s" % annotTETableNameNoRepet
        process = subprocess.Popen(cmd, shell = True)
        self._log.debug("Running : %s" % cmd)
        output = process.communicate()
        self._log.debug("Output:\n%s" % output[0])
        if process.returncode != 0:
            self._logAndRaise("ERROR when launching '%s'" % cmd)
            
        cmd = "sqlCrossMatch.py "
        cmd += "%s " % annotTETableNameNoRepet
        process = subprocess.Popen(cmd, shell = True)
        self._log.debug("Running : %s" % cmd)
        output = process.communicate()
        self._log.debug("Output:\n%s" % output[0])
        if process.returncode != 0:
            self._logAndRaise("ERROR when launching '%s'" % cmd)
            
        self._log.info("FINISHED repeats filtering")
        
        outFileName = "%s_NoRep_%s.align" % (self._projectName, self._extend)
        shutil.move("%s_Palin" % annotTETableNameNoRepet, outFileName)
        os.chdir("..")
        self._log.info("Step 3 finished successfully\n")

class MatcherAndFilterKnownRepeats(SegDupStep):
    
    def __init__(self, configFileName, verbosity = 1):
        SegDupStep.__init__(self, configFileName, verbosity)
        self._sectionName = "matcher"

    def _checkConfig(self):       
        iConfigRules = ConfigRules()
        iConfigRules.addRuleSection(section=self._sectionName, mandatory=True)
        iConfigRules.addRuleOption(section=self._sectionName, option ="Evalue", mandatory=True, type="string")
        iConfigRules.addRuleOption(section="filtering", option ="extend", mandatory=True, type="string")
        iConfigRules.addRuleOption(section="filtering", option ="TE_annotation_table_names", mandatory=True, type="string")
        iConfigRules.addRuleOption(section="filtering", option ="Satellites_annotation_table_names", mandatory=True, type="string")

        iConfigChecker = ConfigChecker(self._configFileName, iConfigRules)
        self._iConfig = iConfigChecker.getConfig()
        self._setAttributesFromConfig()
        
    def _setAttributesFromConfig(self):
        self._projectName = self._iConfig.get("project", "project_name")
        self._projectDir = self._iConfig.get("project", "project_dir")
        self._eValue = self._iConfig.get(self._sectionName, "Evalue")
        self._extend = self._iConfig.get("filtering", "extend")
        self._TETableNames = self._iConfig.get("filtering", "TE_annotation_table_names")
        self._SatTableNames = self._iConfig.get("filtering", "Satellites_annotation_table_names")
        self._doClean = True if self._iConfig.get("matcher", "clean") == "yes" else False
        
    def _checkAttributes(self):
        if not self._TETableNames and not self._SatTableNames:
            self._logAndRaise("ERROR: need to specify at least one annotation table")
        
    def run(self):
        self._log.info("Beginning of step 4")
        self._checkConfig()
        self._checkAttributes()
        stepdir = "%s_Blaster_NoRep_Matcher" % self._projectName
        os.makedirs(stepdir)
        os.chdir(stepdir)
        
        alignFileName = "%s_NoRep_%s.align" % (self._projectName, self._extend)
        alignFilePath = "%s/%s_Blaster_NoRep/%s" % (self._projectDir, self._projectName, alignFileName)
        os.symlink(alignFilePath, alignFileName)
        fastaFileName = "%s.fa" % self._projectName
        os.symlink("%s/%s" % (self._projectDir, fastaFileName), fastaFileName)
        
        cmd = "matcher "
        cmd += "-m %s " % alignFileName
        cmd += "-q %s " % fastaFileName
        cmd += "-E %s " % self._eValue
        cmd += "-j "
        cmd += "-a" 
        process = subprocess.Popen(cmd, shell = True)
        self._log.debug("Running : %s" % cmd)
        output = process.communicate()
        self._log.debug("Output:\n%s" % output[0])
        if process.returncode != 0:
            self._logAndRaise("ERROR when launching '%s'" % cmd)
        
        outMatcherPathFileName = "%s.match.path" % alignFileName
        outMatcherPathTableName = "%s_match_path" % alignFileName.replace('.','_')
        
        iDb = DbFactory.createInstance()
        iDb.createTable(outMatcherPathTableName, "path", outMatcherPathFileName)
        iTpa = TablePathAdaptator(iDb, outMatcherPathTableName)
        outMatcherRangeTableName = outMatcherPathTableName.replace("path", "range")
        iTpa.path2PathRange(outMatcherRangeTableName)        
        
        outMatcherRangeFileName = "%s.match.range" % alignFileName
        
        iDb.exportDataToFile(outMatcherRangeTableName, outMatcherRangeTableName)
        
        shutil.move(outMatcherRangeTableName, outMatcherRangeFileName)

        rangeFH = open(outMatcherRangeFileName, 'r')
        outMatcherAlignFileName = "%s.align" % outMatcherRangeFileName
        alignFH = open(outMatcherAlignFileName, 'w')
        
        line = rangeFH.readline()
        while line:
            alignFH.write("%s\n" % "\t".join(line.split()[1:]))
            line = rangeFH.readline()
        rangeFH.close()
        alignFH.close()
        
        outMatcherAlignTableName =  "%s_align" % outMatcherRangeTableName
        iDb.createTable(outMatcherAlignTableName, "align", outMatcherAlignFileName)

        self._log.info("START repeats filtering")
        
        cmd = "sqlEliminRepeats_Matcher.py "
        cmd += "-b %s " % outMatcherAlignFileName
        if self._TETableNames != "":
            cmd += "-t "
            cmd += " -t ".join(self._TETableNames.split(","))
        if self._SatTableNames != "":
            cmd += " -s "
            cmd += " -s ".join(self._SatTableNames.split(","))
        process = subprocess.Popen(cmd, shell = True)
        self._log.debug("Running : %s" % cmd)
        output = process.communicate()
        self._log.debug("Output:\n%s" % output[0])
        if process.returncode != 0:
            self._logAndRaise("ERROR when launching '%s'" % cmd)
            
        self._log.info("FINISHED repeats filtering")
        
        outMatcherAlignTableNameNoRep = "%s_NoRep" % outMatcherAlignTableName
        iDb.exportDataToFile(outMatcherAlignTableNameNoRep, outMatcherAlignTableNameNoRep)
        
        outFileName = "%s_NoRep_%s_Matcher.align" % (self._projectName, self._extend)
        shutil.move(outMatcherAlignTableNameNoRep, outFileName)
        
        if self._doClean:
            FileUtils.removeFilesByPattern("%s_NoRep_5.align.match*" % self._projectName)
            iDb.dropTable(outMatcherPathTableName)
            iDb.dropTable(outMatcherRangeTableName)
            iDb.dropTable(outMatcherAlignTableName)
        iDb.close()
        os.chdir("..")
        self._log.info("Step 4 finished successfully\n")

class ClusterizeAndRegroupByClusterSize(SegDupStep):
    
    def __init__(self, configFileName, verbosity = 1):
        SegDupStep.__init__(self, configFileName, verbosity)
        self._sectionName = "clustering"

    def _checkConfig(self):       
        iConfigRules = ConfigRules()
        iConfigRules.addRuleSection(section=self._sectionName, mandatory=True)
        iConfigRules.addRuleOption(section=self._sectionName, option ="Evalue", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._sectionName, option ="max_copies_number", mandatory=True, type="integer")
        iConfigRules.addRuleOption(section="filtering", option ="extend", mandatory=True, type="string")

        iConfigChecker = ConfigChecker(self._configFileName, iConfigRules)
        self._iConfig = iConfigChecker.getConfig()
        self._setAttributesFromConfig()
        
    def _setAttributesFromConfig(self):
        self._projectName = self._iConfig.get("project", "project_name")
        self._projectDir = self._iConfig.get("project", "project_dir")
        self._inFastaFilePath = "%s/%s.fa" % (self._projectDir, self._projectName) 
        self._extend = self._iConfig.get("filtering", "extend")
        self._eValue = self._iConfig.get(self._sectionName, "Evalue")
        self._maxCopiesNb = self._iConfig.get(self._sectionName, "max_copies_number")
        self._doClean = True if self._iConfig.get(self._sectionName, "clean") == "yes" else False
        
    def _checkAttributes(self):
        if self._maxCopiesNb < 3:
            self._logAndRaise("ERROR: 'max_copies_number' must be > 2.")
        
    def _createDirAndGFF(self, mapTableName, dirName, mapFileName, setTableName):
        tableFileName = "tables.txt"
        iDb = DbFactory.createInstance()
        if not iDb.isEmpty(mapTableName):
            os.makedirs(dirName)
            os.chdir(dirName)
            iDb.exportDataToFile(mapTableName, mapFileName)
            iDb.convertMapTableIntoSetTable(mapTableName, setTableName)
            with open(tableFileName, 'w') as fH:
                fH.write("SegDupPipeline\tset\t%s\n" % setTableName)
            iGFF3Maker = GFF3Maker(inFastaName = self._genomeTableName, tablesFileName = tableFileName, isWithSequence = False, verbose = self._verbosity)
            iGFF3Maker.run()
            os.chdir("..")
        if self._doClean:
            iDb.dropTable(mapTableName)
        
    def run(self):
        self._log.info("Beginning of step 5")
        self._checkConfig()
        self._checkAttributes()
        stepdir = "%s_Blaster_NoRep_Matcher_Grouper" % self._projectName
        os.makedirs(stepdir)
        os.chdir(stepdir)
        
        alignFileName = "%s_NoRep_%s_Matcher.align" % (self._projectName, self._extend)
        alignFilePath = "%s/%s_Blaster_NoRep_Matcher/%s" % (self._projectDir, self._projectName, alignFileName)
        os.symlink(alignFilePath, alignFileName)
        fastaFileName = "%s.fa" % self._projectName
        os.symlink("%s/%s" % (self._projectDir, fastaFileName), fastaFileName)
        
        cmd = "grouper "
        cmd += "-m %s " % alignFileName
        cmd += "-q %s " % fastaFileName
        cmd += "-s %s " % fastaFileName
        cmd += "-C 0 "
        cmd += "-G -1 "
        cmd += "-E %s " % self._eValue
        process = subprocess.Popen(cmd, shell = True)
        self._log.debug("Running : %s" % cmd)
        output = process.communicate()
        self._log.debug("Output:\n%s" % output[0])
        if process.returncode != 0:
            self._logAndRaise("ERROR when launching '%s'" % cmd)
        
        outGrouperSetFileName = "%s.group.set" % os.path.splitext(fastaFileName)[0]
        shutil.move("%s.group.c0.set" % alignFileName, outGrouperSetFileName)
        outGrouperMapFileName = "%s.map" % os.path.splitext(outGrouperSetFileName)[0]

        SetUtils.convertSetFileIntoMapFile(outGrouperSetFileName, outGrouperMapFileName)
        outGrouperMapTableName = outGrouperMapFileName.replace(".", "_")

        self._log.info("START Segmental Duplication selection")

        cmd = "sqlSelectSegDup.py "
        cmd += "-f %s " % outGrouperMapFileName
        cmd += "-t %s " % outGrouperMapTableName
        cmd += "-x %s " % self._maxCopiesNb
        process = subprocess.Popen(cmd, shell = True)
        self._log.debug("Running : %s" % cmd)
        output = process.communicate()
        self._log.debug("Output:\n%s" % output[0])
        if process.returncode != 0:
            self._logAndRaise("ERROR when launching '%s'" % cmd)
        
        self._log.info("FINISHED Segmental Duplication selection")
        
        map2TableName = "%s_2" % outGrouperMapTableName
        map3TableName = "%s_3to%s" % (outGrouperMapTableName, self._maxCopiesNb)
        mapMoreTableName = "%s_more%s" % (outGrouperMapTableName, self._maxCopiesNb)
        
        set2TableName = map2TableName.replace("map", "set")
        set3TableName = map3TableName.replace("map", "set")
        setMoreTableName = mapMoreTableName.replace("map", "set")
            
        self._genomeTableName = "%s_chr_seq" % self._projectName
        iDb = DbFactory.createInstance()
        iDb.createTable(self._genomeTableName, "seq", self._inFastaFilePath, True)
        
        map2FileName = "%s.group.2.map" % os.path.splitext(fastaFileName)[0]
        self._createDirAndGFF(map2TableName, "clusters2members", map2FileName, set2TableName)
            
        map3FileName = "%s.group.3to%s.map" % (os.path.splitext(fastaFileName)[0], self._maxCopiesNb)
        self._createDirAndGFF(map3TableName, "clusters3To%s" % self._maxCopiesNb, map3FileName, set3TableName)
        
        mapMoreFileName = "%s.group.more%s.map" % (os.path.splitext(fastaFileName)[0], self._maxCopiesNb)
        self._createDirAndGFF(mapMoreTableName, "clustersMore%s" % self._maxCopiesNb, mapMoreFileName, setMoreTableName)
        
        if self._doClean:
            FileUtils.removeFilesByPattern("%s*group*" % self._projectName)
            FileUtils.removeFilesByPattern("%s_NoRep_*_Matcher.align.chains.path*" % self._projectName)
            FileUtils.removeFilesByPattern("clusters*/tables.txt")
            iDb.dropTable(self._genomeTableName)
            iDb.dropTable(outGrouperMapTableName)
        os.chdir("..")
        iDb.close()
        
        self._log.info("Step 5 finished successfully\n")
        
if __name__ == "__main__":
    iSegDupPipeline = SegDupPipeline()
    iSegDupPipeline.setAttributesFromCmdLine()
    iSegDupPipeline.run()      
