#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import os
import ConfigParser
from ConfigParser import MissingSectionHeaderError
from commons.core.utils.RepetOptionParser import RepetOptionParser
from commons.core.utils.FileUtils import FileUtils
from commons.core.sql.DbFactory import DbFactory
from commons.core.sql.TablePathAdaptator import TablePathAdaptator
from commons.core.seq.FastaUtils import FastaUtils
from commons.core.LoggerFactory import LoggerFactory
from commons.core.parsing.FastaParser import FastaParser

LOG_DEPTH = "repet.tools"

## Splices annotations from genome
#
class SplicerFromAnnotation(object):
    
    def __init__(self, inputFileName = "", configFileName = "", copyType = 1, identity = 80, outputFile = "", minLength = 2000, verbose = 3):
        self.setInputFileName(inputFileName)
        self.setConfigFileName(configFileName)
        self._copyType = copyType
        self._identity = identity
        self.setOutputFileName(outputFile)
        self._minLength = minLength
        self._verbosity = verbose
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbosity)
        
    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise Exception(errorMsg)
    
    def setAttributesFromCmdLine(self):
        usage = "SplicerFromAnnotation.py [options]"
        desc = "To splice annotations from genome. These annotations are Full Length Copies or Full Length Fragments according to consensus (or TEs) library.\n"
        desc += "The consensus (or TEs) library and annotations information are necessary, from TEannot pipeline for example.\n"
        desc += "Connection parameters to database  are required from the provided configuration file.\n"
        examples = "Example : with a project called 'MyTEannotAnalysis', MyTEannotAnalysis_chr_allTEs_nr_noSSR_join_path and MyTEannotAnalysis_refTEs_seq must exit:\n"
        examples += "\t$ SplicerFromAnnotation.py -i MyTEannotAnalysis.fa -C config.cfg -t 1\n"
        examples += "\t Results are in MyTEannotAnalysis.fa.splice file, containing the spliced sequences whose length is greater than the 'minLength'\n"
        parser = RepetOptionParser(description = desc, epilog = examples, usage = usage)
        parser.add_option("-i", "--file",       dest = "inputFastaFileName", action = "store", type = "string", help = "input file [mandatory]", default = "")
        parser.add_option("-C", "--config",     dest = "configFileName",     action = "store", type = "string", help = "configuration file name to set database connection [mandatory]", default = "")
        parser.add_option("-t", "--copyType",   dest = "copyType",           action = "store", type = "int",    help = "1 is Full Length Copy [default: 1], 2 is Full Length Fragment", default = 1)
        parser.add_option("-I", "--identity",   dest = "identity",           action = "store", type = "float",  help = "minimal copy identity percent (between 0 and 100) [default: 80]", default = 80)
        parser.add_option("-o", "--outputFile", dest = "outputFile",         action = "store", type = "string", help = "output fasta file [default=<input>.splice]", default = "")
        parser.add_option("-m", "--minLength",  dest = "minLength",          action = "store", type = "int",    help = "minimal sequence length in the spliced fasta file [default: 2000]", default = 2000)
        parser.add_option("-v", "--verbose",    dest = "verbose",            action = "store", type = "int",    help = "verbosity level [default: 3]", default = 3)
        options = parser.parse_args()[0]
        self._setAttributesFromOptions(options)
        
    def _setAttributesFromOptions(self, options):
        self.setInputFileName(options.inputFastaFileName)
        self.setConfigFileName(options.configFileName)
        self.setCopyType(options.copyType)
        self.setIdentity(options.identity)
        self.setOutputFileName(options.outputFile)
        self.setMinLength(options.minLength)
        self.setVerbose(options.verbose)
        
    def setInputFileName(self, inputFastaFileName):
        self._inputFastaFileName = ""
        if inputFastaFileName:
            self._inputFastaFileName = inputFastaFileName
            self._projectName = os.path.splitext(os.path.basename(self._inputFastaFileName))[0]
            self.pathTable = "%s_chr_allTEs_nr_noSSR_join_path" % self._projectName
            self.refTEsSeqTable = "%s_refTEs_seq" % self._projectName

            iFP = FastaParser(self._inputFastaFileName)
            iFP.getInfos()
            self.genomeSize = iFP.getSize()
            self.nbSeqGenome = iFP.getNbSequences()
            
    def setConfigFileName(self, configFileName):
        self._configFileName = ""
        if configFileName:
            self._configFileName = configFileName
            config = ConfigParser.ConfigParser()

            try :
                config.readfp(open(self._configFileName))
            except MissingSectionHeaderError:
                self._logAndRaise("Configuration file '" + self._configFileName + "' must begin with a section header")
    
            self.setup_env(config)
            
    def setCopyType(self, copyType):
        self._copyType = copyType
        
    def setIdentity(self,identity):
        self._identity = identity
        
    def setOutputFileName(self, outputFile):
        if outputFile == "":
            self._outputFileName = "%s.splice" % self._inputFastaFileName
        else:
            self._outputFileName = outputFile
        
    def setMinLength(self, minLength):
        self._minLength = minLength

    def setVerbose(self, verbose):
        self._verbosity = verbose
        
    def setup_env(self, config):
        os.environ["REPET_HOST"] = config.get("repet_env", "repet_host")
        os.environ["REPET_USER"] = config.get("repet_env", "repet_user")
        os.environ["REPET_PW"] = config.get("repet_env", "repet_pw")
        os.environ["REPET_DB"] = config.get("repet_env", "repet_db")
        os.environ["REPET_PORT"] = config.get("repet_env", "repet_port")

    def checkOptions(self):
        if self._inputFastaFileName:
            if not FileUtils.isRessourceExists(self._inputFastaFileName):
                self._logAndRaise("Input fasta file '' does not exist!" % self._inputFastaFileName)
        else:
            self._logAndRaise("No specified mandatory '-i' option!")
       
        if self._copyType not in [1, 2]:
            self._logAndRaise("Copy type must be only 1 or 2!")
            
        if self._identity < 0 or self._identity > 100:
            self._logAndRaise("Identity must be in [0:100]!")
            
        if self._configFileName != "":
            self._iDb = DbFactory.createInstance(self._configFileName)
        else:
            self._logAndRaise("No specified configuration file!")

        if not self._iDb.doesTableExist(self.pathTable):
            self._logAndRaise("Annotation table  %s doesn't exist!" % self.pathTable)
        if not self._iDb.doesTableExist(self.refTEsSeqTable):
            self._logAndRaise("sequences table  %s doesn't exist!" % self.refTEsSeqTable)
                    
    def run(self):
        LoggerFactory.setLevel(self._log, self._verbosity)
        self.checkOptions()
        
        self._log.info("START SplicerFromAnnotation")
        self._log.debug("Input info file: %s" % self._inputFastaFileName)
        self._log.debug("Copy type is: %s" % self._copyType)
        self._log.debug("Identity is: %s" % self._identity)
        self._log.debug("Host is: %s" % os.environ["REPET_HOST"])
        self._log.debug("User is: %s" % os.environ["REPET_USER"])
        self._log.debug("Database is: %s" % os.environ["REPET_DB"])
        self._log.debug("Database port is: %s" % os.environ["REPET_PORT"])
        
        cmd = "PostAnalyzeTELib.py -a 3 -p %s -s %s -g %s" % (self.pathTable, self.refTEsSeqTable, self.genomeSize)
        os.system(cmd)
        
        cmd = "GetSpecificTELibAccordingToAnnotation.py -i %s_chr_allTEs_nr_noSSR_join_path.annotStatsPerTE.tab -t %s_refTEs_seq -v 2" % (self._projectName, self._projectName)
        os.system(cmd) 

        copyType = "Frag"
        if self._copyType == 1:
            copyType = "Copy"
        lines = open("%s.annotStatsPerTE_FullLength%s.txt" % (self.pathTable, copyType), "r").readlines()[1:]

        if not lines:
            self._log.warning("There is no consensus in this copy type")
        else:
            self._log.info("Selection of annotations which will be spliced")

            lConsensusHeader_copyType = [i.split('\t', 1)[0] for i in lines]

            annotationTableSupIdentity = "%s_annotationIdentitySup%i_path" % (self._projectName, int(self._identity))
            self._iDb.copyTable(self.pathTable, annotationTableSupIdentity, params = "where identity >= %f" % self._identity)

            iTPA = TablePathAdaptator(self._iDb, annotationTableSupIdentity)
            lAllDistinctPath = []
            for consensusName in lConsensusHeader_copyType:
                lDistinctPath = iTPA.getIdListFromSubject(consensusName)
                lAllDistinctPath.extend(lDistinctPath)
            lAllDistinctPath.sort()
                
            annotationTableToBeSpliced = "%s_annotationToSplice_path" % self._projectName
            lAllDistinctPath = [str(pathID) for pathID in lAllDistinctPath]
            self._iDb.copyTable(self.pathTable, annotationTableToBeSpliced, params = "where path IN (%s)" % ", ".join(lAllDistinctPath))

            self._iDb.close()
            self._log.info("Splicing")
            
            cmd = "SpliceTEsFromGenome.py -i %s -f path -g %s -o %s -C %s -v 2" % (annotationTableToBeSpliced, self._inputFastaFileName, self._outputFileName, self._configFileName)
            os.system(cmd)

            self._log.info("Keep only sequences which length > %s" % self._minLength)
            FastaUtils.dbLengthFilter(self._minLength, self._outputFileName, 1)
            os.remove("%s.Inf%i" % (self._outputFileName, self._minLength))
            os.remove("%s" % self._outputFileName)
            os.rename("%s.Sup%i" % (self._outputFileName, self._minLength), self._outputFileName)
        
        self._log.info("END SplicerFromAnnotation")
        return 0

if __name__ == '__main__':
    iSFA = SplicerFromAnnotation()
    iSFA.setAttributesFromCmdLine()
    iSFA.run()