#!/usr/bin/env python

from commons.core.utils.RepetOptionParser import RepetOptionParser
from commons.core.utils.FileUtils import FileUtils
import sys

"""
This script give statistics for the result of PASTEC.
usage: python StatPastec.py [options]
    --help: this help"
    --i: name of the input file, which is PASTEC type file.(format='OutputFile.classif')"
    --o: name of the output file (default=inFileName+'_statsPASTEC.txt')"
    --v: verbose (default=0/1)"
"""

CLASS_I = 'I'
CLASS_II = 'II'
COLUMNS_NB = 8
CLASS_COLUMN = 4
ORDER_COLUMN = 5
CHIMERIC = " potential chimeric*"
DWICKERCODE = {
                "I" : "RXX", 
                "II" : "DXX", 
                "LTR" : "RLX", 
                "DIRS" : "RYX", 
                "PLE" : "RPX", 
                "LINE" : "RIX", 
                "SINE" : "RSX", 
                "TIR" : "DTX", 
                "Crypton" : "DYX", 
                "Helitron" : "DHX", 
                "Maverick" : "DMX", 
                "MITE" : "DXX-MITE", 
                "LARD" : "RXX-LARD", 
                "TRIM" : "RXX-TRIM",
               }
DORDER2CLASS = {
                    "LTR": CLASS_I,
                    "DIRS": CLASS_I,
                    "PLE": CLASS_I,
                    "LINE": CLASS_I,
                    "SINE": CLASS_I,
                    "LARD": CLASS_I,
                    "TRIM": CLASS_I,
                    "TIR": CLASS_II,
                    "Crypton": CLASS_II,
                    "Maverick": CLASS_II,
                    "Helitron": CLASS_II,
                    "MITE" : CLASS_II
                    }

class StatPastec(object):
    
    def __init__(self, inFileName = "", outFileName = None):
        self._inFileName = inFileName
        self.setOutFileName(outFileName)
        self._verbose = 0
        self._dStat = {}

        self._nbElementClassI = 0
        self._nbElementClassII = 0
        self._nbElementPC = 0
        self._nbElementRDNA = 0
        self._nbElementSSR = 0

        self._nbElementsChimeric = 0
        self._nbElementNoCatNoCat = 0
        self._nbElementNoCatSeveralOrders = 0
        self._nbElementSeveralOrders = 0
        self._nbElementNASeveralOrders = 0
                
    def setAttributesFromCmdLine(self):
        description = "This script give statistics for the result of PASTEC, by default from file with .classif extension.\n"
        epilog = "Example 1: common usage\n"
        epilog += "\t$ StatPastec.py --infile outfile.classif\n"
        epilog += "Example 2: launch with stats output file name\n"
        epilog += "\t$ StatPastec.py -i outfile.classif -o statClassif.txt\n"
        parser = RepetOptionParser(description = description, epilog = epilog)
        parser.add_option('-i', '--infile', dest='inFileName', help='name of the input file')
        parser.add_option('-o', '--outfile', dest='outFileName', help='name of the output file', default = None)
        parser.add_option('-v', '--verbosity', dest='verbose', help='verbosity default = 0, else 1', default = 0, type = "int")
        options = parser.parse_args()[0]
        self.setAttributeFromOptions(options)
        
    def setAttributeFromOptions(self, options):
        self.setInputFileName(options.inFileName)
        self.setOutFileName(options.outFileName)
        self.setVerbose(options.verbose)
    
    def setInputFileName(self, inFileName):
        self._inFileName = inFileName
        
    def setOutFileName(self, outFileName):
        if outFileName != None:
            self._outFileName = outFileName
        else:
            self._outFileName = self._inFileName + '_stats.txt'
        
    def setVerbose(self, verbose):
        self._verbose = verbose
         
    def checkOptions(self):
        if self._inFileName == "":
            raise Exception("ERROR: No input file specified!")
        elif not FileUtils.isRessourceExists(self._inFileName):
            raise Exception("ERROR: This input file specified doesn't exist!")
                    
    def run(self):
        if self._verbose > 0:
            print "START %s" % ( sys.argv[0].split("/")[-1] )
            sys.stdout.flush()
            
        self.checkOptions()
        aClassifFile = open(self._inFileName,"r")
        aClassifLine = aClassifFile.readline()
        lineNumber = 0
        totalElement = 0 
        while aClassifLine :
            lineNumber += 1
            lFieldsOfClassifLine = aClassifLine.split("\t")
            if len(lFieldsOfClassifLine) == COLUMNS_NB:
                totalElement += 1 
                self.incrementeForOrder(lFieldsOfClassifLine)
                self.incrementeForSeveralOrders(lFieldsOfClassifLine)
                self.incrementeForNoCat(lFieldsOfClassifLine)
            elif self._verbose > 0:
                print "line %s has not %s columns" % (lineNumber, COLUMNS_NB)
            aClassifLine = aClassifFile.readline()
        self._writeStat(totalElement)

        if self._verbose > 0:
            print "END %s" % ( sys.argv[0].split("/")[-1] )
            sys.stdout.flush()
        
    def incrementeForOrder(self, lFieldsOfClassifLine):
        order = lFieldsOfClassifLine[ORDER_COLUMN]
        if order in DORDER2CLASS.keys():
            self._incrementAStatKeyFound("%s total (%s)" % (order, DWICKERCODE[order]))
            self._incrementAStatKeyFound("Class%s + one order" % DORDER2CLASS[order])
            self._incrementAStatKeyFound("Class%s total (%s)" % (DORDER2CLASS[order], DWICKERCODE[DORDER2CLASS[order]]))
            if "complete" in lFieldsOfClassifLine:
                self._incrementAStatKeyFound("%s comp" % order)
            elif "incomplete" in lFieldsOfClassifLine:
                self._incrementAStatKeyFound("%s incomp" % order)
            if "PotentialChimeric" in lFieldsOfClassifLine:
                self._nbElementsChimeric += 1
                self._incrementAStatKeyFound("Class%s%s" % (DORDER2CLASS[order], CHIMERIC))
                self._incrementAStatKeyFound("".join([order, CHIMERIC]))
        elif order in ["SSR", "rDNA", "PotentialHostGene"]:
            self._incrementAStatKeyFound("%s total" % order)
            if "PotentialChimeric" in lFieldsOfClassifLine:
                self._nbElementsChimeric += 1
                self._incrementAStatKeyFound("".join([order, CHIMERIC]))
        
    def incrementeForNoCat(self, lFieldsOfClassifLine):
        order = str(lFieldsOfClassifLine[ORDER_COLUMN])
        classField = str(lFieldsOfClassifLine[CLASS_COLUMN])
        if order == "noCat" and classField == "noCat":
            self._nbElementNoCatNoCat += 1
        if order == "noCat" and (classField == "I" or classField == "II"):
            self._incrementAStatKeyFound("Class%s + noCat order" % classField)
            self._incrementAStatKeyFound("Class%s total (%s)" % (classField, DWICKERCODE[classField]))
        
    def incrementeForSeveralOrders(self, lFieldsOfClassifLine):
        order = str(lFieldsOfClassifLine[ORDER_COLUMN])
        classField = str(lFieldsOfClassifLine[CLASS_COLUMN])
        if "|" in order:
            self._nbElementsChimeric += 1
            self._nbElementSeveralOrders += 1
            if classField == "noCat":
                self._nbElementNoCatSeveralOrders += 1
            elif classField == "NA":
                self._incrementAStatKeyFound("%s%s total" % (classField, CHIMERIC))
                self._nbElementNASeveralOrders += 1
            else:
                self._incrementAStatKeyFound("Class%s + several orders" % classField)
                self._incrementAStatKeyFound("Class%s%s" % (classField, CHIMERIC))
                self._incrementAStatKeyFound("Class%s total (%s)" % (classField, DWICKERCODE[classField]))
#            elif classField != "NA":
#                self._incrementAStatKeyFound("Class%s + several orders" % classField)
#                self._incrementAStatKeyFound("Class%s%s" % (classField, CHIMERIC))
#                self._incrementAStatKeyFound("Class%s total (%s)" % (classField, DWICKERCODE[classField]))
#            else:
#                self._incrementAStatKeyFound("%s%s" % (classField, CHIMERIC))
#                self._nbElementNASeveralOrders += 1
                
                
    def groupList(self, lStatKey):
        lStatKeyClassI = []
        lStatKeyClassII = []
        lStatKeyOrderClassI = []
        lStatKeyOrderClassII = []
        lStatKeyNotAvailable = []
        for aStatKeyName in lStatKey :
            lOrderKeyName = aStatKeyName.split(" ")
            aOrderKeyName = lOrderKeyName[0]
            if DORDER2CLASS.has_key(aOrderKeyName) :
                aClassName = DORDER2CLASS[aOrderKeyName]
                if aClassName == CLASS_I :
                    lStatKeyOrderClassI.append(aStatKeyName)
                elif aClassName == CLASS_II :
                    lStatKeyOrderClassII.append(aStatKeyName)
            elif aOrderKeyName == "ClassI":
                lStatKeyClassI.append(aStatKeyName)
            elif aOrderKeyName == "ClassII":
                lStatKeyClassII.append(aStatKeyName)
            else:
                lStatKeyNotAvailable.append(aStatKeyName)
        lStatKeyClassI.sort()
        self._nbElementClassI = lStatKeyClassI
        lStatKeyClassII.sort()
        lStatKeyOrderClassI.sort()
        lStatKeyOrderClassII.sort()
        lStatKeyNotAvailable.sort()
        return lStatKeyClassI, lStatKeyClassII, lStatKeyOrderClassI, lStatKeyOrderClassII, lStatKeyNotAvailable
  
    def _incrementAStatKeyFound(self, aKeyName):
        if self._dStat.has_key(aKeyName):
            self._dStat[aKeyName] += 1 
        else:
            self._dStat[aKeyName] = 1

    def _writeStat(self, total):
        outStatFile = open(self._outFileName,"w")
        lStatKey = self._dStat.keys()
        lStatKeyClassI, lStatKeyClassII, lStatKeyOrderClassI, lStatKeyOrderClassII, lStatKeyNotAvailable = self.groupList(lStatKey)
            
        for aStatKeyName in lStatKeyOrderClassI:
            self._writeALineInFile(outStatFile, "%s: %d (%.2f%%)" % (aStatKeyName, self._dStat[aStatKeyName], 100*self._dStat[aStatKeyName]/float(total)))
        self._writeALineInFile(outStatFile, "")
                
        for aStatKeyName in lStatKeyClassI:
            self._writeALineInFile(outStatFile, "%s: %d (%.2f%%)" % (aStatKeyName, self._dStat[aStatKeyName], 100*self._dStat[aStatKeyName]/float(total)))
        self._writeALineInFile(outStatFile, "")
            
        for aStatKeyName in lStatKeyOrderClassII:
            self._writeALineInFile(outStatFile, "%s: %d (%.2f%%)" % (aStatKeyName, self._dStat[aStatKeyName], 100*self._dStat[aStatKeyName]/float(total)))
        self._writeALineInFile(outStatFile, "")
                
        for aStatKeyName in lStatKeyClassII:
            self._writeALineInFile(outStatFile, "%s: %d (%.2f%%)" % (aStatKeyName, self._dStat[aStatKeyName], 100*self._dStat[aStatKeyName]/float(total)))
        self._writeALineInFile(outStatFile, "")
            
        for aStatKeyName in lStatKeyNotAvailable:
            self._writeALineInFile(outStatFile, "%s: %d (%.2f%%)" % (aStatKeyName, self._dStat[aStatKeyName], 100*self._dStat[aStatKeyName]/float(total)))
        self._writeALineInFile(outStatFile, "")
        
        if self._nbElementNoCatSeveralOrders != 0:
            self._writeALineInFile(outStatFile, "Nb noCat class with several orders (XXX): %i (%.2f%%)" % (self._nbElementNoCatSeveralOrders, 100*self._nbElementNoCatSeveralOrders/float(total)))
        if self._nbElementNASeveralOrders != 0:
            self._writeALineInFile(outStatFile, "Nb NA with several orders (NA): %i (%.2f%%)" % (self._nbElementNASeveralOrders, 100*self._nbElementNASeveralOrders/float(total)))
        if self._nbElementSeveralOrders != 0:
            self._writeALineInFile(outStatFile, "Nb elements with several orders: %i (%.2f%%)" % (self._nbElementSeveralOrders, 100*self._nbElementSeveralOrders/float(total)))
        if self._nbElementsChimeric != 0:
            self._writeALineInFile(outStatFile, "Nb Potential chimeric*: %i (%.2f%%)" % (self._nbElementsChimeric, 100*self._nbElementsChimeric/float(total)))
        self._writeALineInFile(outStatFile, "")
        
        if self._nbElementNoCatNoCat != 0:
            self._writeALineInFile(outStatFile, "Nb noCat at class and order levels (noCat): %i (%.2f%%)" % (self._nbElementNoCatNoCat, 100*self._nbElementNoCatNoCat/float(total)))
        self._writeALineInFile(outStatFile, "")

        self._writeALineInFile(outStatFile, "\t-------------------------Summary--------------------------------")
        self._writeALineInFile(outStatFile, "")

        if lStatKeyClassI != []:
            self._nbElementClassI = self._dStat[lStatKeyClassI[-1]]
            self._writeALineInFile(outStatFile, "RXX: %d (%.2f%%)" % (self._nbElementClassI,100*self._nbElementClassI/float(total)))
        if lStatKeyClassII != []:
            self._nbElementClassII = self._dStat[lStatKeyClassII[-1]]
            self._writeALineInFile(outStatFile, "DXX: %d (%.2f%%)" % (self._nbElementClassII,100*self._nbElementClassII/float(total)))
        if self._nbElementNoCatSeveralOrders != 0:
            self._writeALineInFile(outStatFile, "XXX: %i (%.2f%%)" % (self._nbElementNoCatSeveralOrders, 100*self._nbElementNoCatSeveralOrders/float(total)))
#        if self._nbElementNASeveralOrders != 0:
#            self._writeALineInFile(outStatFile, "NA: %i (%.2f%%)" % (self._nbElementNASeveralOrders, 100*self._nbElementNASeveralOrders/float(total)))
        l = len(lStatKeyNotAvailable) - 1
        while l != -1:
            label = str(lStatKeyNotAvailable[l-1])
            if "total" in label:
                nbElements = self._dStat[label]
                label = label.split(" total")[0]
                self._writeALineInFile(outStatFile, "%s: %i (%.2f%%)" % (label, nbElements, 100*nbElements/float(total)))
            l -= 1
#        if len(lStatKeyNotAvailable) >= 1:
#            self._nbElementPH = self._dStat[lStatKeyNotAvailable[0]]
#            self._writeALineInFile(outStatFile, "Potential HostGene: %i (%.2f%%)" % (self._nbElementPH, 100*self._nbElementPH/float(total)))
#            if len(lStatKeyNotAvailable) >= 2:
#                self._nbElementSSR =self._dStat[lStatKeyNotAvailable[1]]
#                self._writeALineInFile(outStatFile, "SSR: %i (%.2f%%)" % (self._nbElementSSR, 100*self._nbElementSSR/float(total)))
#                if len(lStatKeyNotAvailable) >= 3:
#                    self._nbElementRDNA =self._dStat[lStatKeyNotAvailable[2]]
#                    self._writeALineInFile(outStatFile, "rDNA: %i (%.2f%%)" % (self._nbElementRDNA, 100*self._nbElementRDNA/float(total)))
        if self._nbElementNoCatNoCat != 0:
            self._writeALineInFile(outStatFile, "noCat: %i (%.2f%%)" % (self._nbElementNoCatNoCat, 100*self._nbElementNoCatNoCat/float(total)))
        self._writeALineInFile(outStatFile, "TOTAL: %i (%.2f%%)" % (total, 100*total/float(total)))
        self._writeALineInFile(outStatFile, "")
        
        self._writeNotes(outStatFile)
        outStatFile.close()

    def _writeNotes(self, outStatFile):
        self._writeALineInFile(outStatFile, "\t-----------------------------NOTES------------------------------")
        self._writeALineInFile(outStatFile, "")
        self._writeALineInFile(outStatFile, "Read  A unified classification system for eukaryotic transposable elements. Wicker T. et al 2007, Nat Rev Genet. 8:973-982")
        self._writeALineInFile(outStatFile, "1) Potential chimeric is a tag different from completeness. So in order to have a total of classification type,")
        self._writeALineInFile(outStatFile, "\tthe number of complete and incomplete are added, but not the number of potential chimeric.")
        self._writeALineInFile(outStatFile, "2) Consensus with several orders are potential chimeric, but no classification is better than")
        self._writeALineInFile(outStatFile, "\tthe others (same confidence index between all classifications found)")
        self._writeALineInFile(outStatFile, "3) Following all classifications that can be found, with Wicker's code:")
        self._writeALineInFile(outStatFile, "")
        self._writeALineInFile(outStatFile, "Transposable elements :")
        self._writeALineInFile(outStatFile, "\tClass I (RXX)")
        for key, value in sorted(DORDER2CLASS.iteritems()):
            if value == CLASS_I:
                self._writeALineInFile(outStatFile, "\t\t%s (%s)" % (key, DWICKERCODE[key]))
        
        self._writeALineInFile(outStatFile, "")
        self._writeALineInFile(outStatFile, "\tClass II (DXX)")
        for key, value in sorted(DORDER2CLASS.iteritems()):
            if value == CLASS_II:
                self._writeALineInFile(outStatFile, "\t\t%s (%s)" % (key, DWICKERCODE[key]))
        self._writeALineInFile(outStatFile, "\tnoCat (sequence not classified at class AND order levels) ")
        self._writeALineInFile(outStatFile, "\tXXX (sequence not classified at class level and with potential several orders)") 
        
        self._writeALineInFile(outStatFile, "")
        self._writeALineInFile(outStatFile, "Not transposable elements :")
        self._writeALineInFile(outStatFile, "\tPotentialHostGene\n\trDNA\n\tSSR")
        self._writeALineInFile(outStatFile, "")
        self._writeALineInFile(outStatFile, "If a classification is not in the statistical results, it means that none of the analyzed consensus has been classified in this way.")
 
    def _writeALineInFile(self, outStatFile, aLine):
        outStatFile.write(aLine + "\n")
        if self._verbose > 0:
            print (aLine)
DWICKERCODE                    
if __name__ == '__main__':
    launcher = StatPastec()
    launcher.setAttributesFromCmdLine()
    launcher.run()