#!/usr/bin/env python

import os, sys, getopt, exceptions, logging

import pyRepet.seq.BioseqDB
import pyRepet.coord.Map
import pyRepet.util.Stat

#-----------------------------------------------------------------------------

def help():

    """
    Give the list of the command-line options.
    """

    print
    print "usage:",sys.argv[0]," [ options ]"
    print "options:"
    print "     -h: this help"
    print "     -f: name of the fasta file, with the consensus"
    print "     -m: name of the map file (output from Grouper)"
    print "output on stdout + 1 txt file"
    print

#-----------------------------------------------------------------------------

def loadInputFastaFile( fastaFileName, verbose=0 ):

    string = "load the input fasta file '%s'..." % ( fastaFileName )
    if verbose > 0:
        print string; sys.stdout.flush()
    bsDB = pyRepet.seq.BioseqDB.BioseqDB( fastaFileName )
    tmpString = "nb of sequences: %i" % ( bsDB.getSize() )
    string += "\n%s" % ( tmpString )
    if verbose > 0:
        print tmpString; sys.stdout.flush()

    logging.info(string)

    return bsDB

#-----------------------------------------------------------------------------

def loadInputMapFile( mapFileName, verbose=0 ):

    string = "load the input map file '%s'..." % ( mapFileName )
    if verbose > 0:
        print string; sys.stdout.flush()

    clustMethod = "Grouper"
    mapFile = open( mapFileName, "r" )
    line = mapFile.readline()
    count = 1
    dGr2Mb = {}
    dGr2Seq = {}
    dCl2Mb = {}
    dCl2Gr = {}
    dSeq2Mb = {}
    dSeq2Gr = {}
    dSeq2Cl = {}

    # for each member in the input file
    while True:
        if line == "":
            break

        newMap = pyRepet.coord.Map.Map()
        data = line.split("\t")
        name = data[0]
        seqname = data[1]
        start = data[2]
        end = data[3][:-1]
        t = ( name, seqname, start, end )
        newMap.set_from_tuple( t, count )

        # retrieve clusterID, groupID and memberID
        clusterID = name.split("Cl")[1]
        groupID = name.split("Cl")[0].split("Gr")[1]
        if "Q" in name.split("Gr")[0]:
            memberID = name.split("Gr")[0].split("MbQ")[1]
        elif "S" in name:
            memberID = name.split("Gr")[0].split("MbS")[1]

        # record the data per group
        if dGr2Mb.has_key( groupID ):
            dGr2Mb[ groupID ].append( newMap )
        else:
            dGr2Mb[ groupID ] = [ newMap ]
        if dGr2Seq.has_key( groupID ):
            if seqname not in dGr2Seq[groupID]:
                dGr2Seq[groupID].append( seqname )
        else:
            dGr2Seq[groupID] = [ seqname ]

        # record the data per cluster
        if dCl2Mb.has_key( clusterID ):
            dCl2Mb[ clusterID ].append( newMap )
        else:
            dCl2Mb[ clusterID ] = [ newMap ]
        if dCl2Gr.has_key( clusterID ):
            if groupID not in dCl2Gr[ clusterID ]:
                dCl2Gr[ clusterID ].append( groupID )
        else:
            dCl2Gr[ clusterID ] = [ groupID ]

        # record the data per sequence
        if dSeq2Mb.has_key( seqname ):
            dSeq2Mb[ seqname ].append( newMap )
        else:
            dSeq2Mb[ seqname ] = [ newMap ]
        if dSeq2Gr.has_key( seqname ):
            if groupID not in dSeq2Gr[ seqname ]:
                dSeq2Gr[ seqname ].append( groupID )
        else:
            dSeq2Gr[ seqname ] = [ groupID ]
        if dSeq2Cl.has_key( seqname ):
            if clusterID not in dSeq2Cl[ seqname ]:
                dSeq2Cl[ seqname ].append( clusterID )
        else:
            dSeq2Cl[ seqname ] = [ clusterID ]

        line = mapFile.readline()
        count += 1

    mapFile.close()

    tmpString = "nb of groups: %i" % ( len(dGr2Mb.keys()) )
    string += "\n%s" % ( tmpString )
    if verbose > 0:
        print tmpString; sys.stdout.flush()

    tmpString = "nb of clusters: %i" % ( len(dCl2Mb.keys()) )
    string += "\n%s" % ( tmpString )
    if verbose > 0:
        print tmpString; sys.stdout.flush()

    tmpString = "nb of sequences (not members!): %i" % ( len(dSeq2Mb.keys()) )
    string += "\n%s" % ( tmpString )
    if verbose > 0:
        print tmpString; sys.stdout.flush()

    logging.info(string)
    
    return dGr2Mb, dGr2Seq, dCl2Mb, dCl2Gr, dSeq2Mb, dSeq2Gr, dSeq2Cl

#-----------------------------------------------------------------------------

def getSummaryStats( bsDB, mapFileName, dGr2Mb, dGr2Seq, dCl2Mb, dCl2Gr, dSeq2Mb, dSeq2Gr, dSeq2Cl, verbose=0 ):

    string = "summary statistics:\n"
    if verbose > 0: print string

    stat = pyRepet.util.Stat.Stat()

    tmpString = "how many members per group ?"
    string += "\n%s" % ( tmpString )
    if verbose > 0:
        print tmpString; sys.stdout.flush()
    for groupID in dGr2Mb.keys():
        stat.add( len(dGr2Mb[groupID]) )
    tmpString = stat.string()
    string += "\n%s" % ( tmpString )
    if verbose > 0:
        print tmpString; sys.stdout.flush()
    tmpString = stat.stringQuantiles()
    string += "\n%s" % ( tmpString )
    if verbose > 0:
        print tmpString; sys.stdout.flush()
    tmpString = "Q90: %i" % ( stat.quantile(0.9) )
    tmpString += "\nQ95: %i" % ( stat.quantile(0.95) )
    tmpString += "\nQ99: %i" % ( stat.quantile(0.99) )
    tmpString += "\nQ99.5: %i" % ( stat.quantile(0.995) )
    tmpString += "\nQ99.9: %i" % ( stat.quantile(0.999) )
    string += "\n%s" % ( tmpString )
    if verbose > 0:
        print tmpString; sys.stdout.flush()

    stat.reset()

    tmpString = "\nhow many sequences per group ?"
    string += "\n%s" % ( tmpString )
    if verbose > 0:
        print tmpString; sys.stdout.flush()
    for groupID in dGr2Seq.keys():
        stat.add( len(dGr2Seq[groupID]) )
    tmpString = stat.string()
    string += "\n%s" % ( tmpString )
    if verbose > 0:
        print tmpString; sys.stdout.flush()
    tmpString = stat.stringQuantiles()
    string += "\n%s" % ( tmpString )
    if verbose > 0:
        print tmpString; sys.stdout.flush()
    tmpString = "Q90: %i" % ( stat.quantile(0.9) )
    tmpString += "\nQ95: %i" % ( stat.quantile(0.95) )
    tmpString += "\nQ99: %i" % ( stat.quantile(0.99) )
    tmpString += "\nQ99.5: %i" % ( stat.quantile(0.995) )
    tmpString += "\nQ99.9: %i" % ( stat.quantile(0.999) )
    string += "\n%s" % ( tmpString )
    if verbose > 0:
        print tmpString; sys.stdout.flush()

    stat.reset()

    tmpString = "\nhow many groups per cluster ?"
    string += "\n%s" % ( tmpString )
    if verbose > 0:
        print tmpString; sys.stdout.flush()
    for clusterID in dCl2Gr.keys():
        stat.add( len(dCl2Gr[clusterID]) )
    tmpString = stat.string()
    string += "\n%s" % ( tmpString )
    if verbose > 0:
        print tmpString; sys.stdout.flush()
    tmpString = stat.stringQuantiles()
    string += "\n%s" % ( tmpString )
    if verbose > 0:
        print tmpString; sys.stdout.flush()
    tmpString = "Q90: %i" % ( stat.quantile(0.9) )
    tmpString += "\nQ95: %i" % ( stat.quantile(0.95) )
    tmpString += "\nQ99: %i" % ( stat.quantile(0.99) )
    string += "\n%s" % ( tmpString )
    if verbose > 0:
        print tmpString; sys.stdout.flush()

    stat.reset()

    tmpString = "\nare sequences in several groups ?"
    string += "\n%s" % ( tmpString )
    if verbose > 0:
        print tmpString; sys.stdout.flush()
    for seqID in dSeq2Gr.keys():
        stat.add( len(dSeq2Gr[seqID]) )
    tmpString = stat.string()
    string += "\n%s" % ( tmpString )
    if verbose > 0:
        print tmpString; sys.stdout.flush()
    tmpString = stat.stringQuantiles()
    string += "\n%s" % ( tmpString )
    if verbose > 0:
        print tmpString; sys.stdout.flush()
    tmpString = "Q90: %i" % ( stat.quantile(0.9) )
    tmpString += "\nQ95: %i" % ( stat.quantile(0.95) )
    tmpString += "\nQ99: %i" % ( stat.quantile(0.99) )
    string += "\n%s" % ( tmpString )
    if verbose > 0:
        print tmpString; sys.stdout.flush()

    stat.reset()

    tmpString = "\nsequences not present in any group: %i (%.2f%%)" % ( bsDB.getSize() - len(dSeq2Mb.keys()), 100*(bsDB.getSize() - len(dSeq2Mb.keys()))/float(bsDB.getSize()) )
    string += "\n%s" % ( tmpString )
    if verbose > 0:
        print tmpString; sys.stdout.flush()
    tmpString = ""
    for bs in bsDB.db:
        if not dSeq2Mb.has_key( bs.header ):
            tmpString +=  " %s\n" % ( bs.header )
    string += "\n%s" % ( tmpString )
    if verbose > 0:
        print tmpString; sys.stdout.flush()

    outFileName = "%s_distribGroupSize.txt" % ( mapFileName )
    outFile = open( outFileName, "w" )
    outFile.write( "groupID\tnbMembers\tnbSequences\n" )
    for groupID in dGr2Mb.keys():
        lSeq = []
        for iMap in dGr2Mb[groupID]:
            if iMap.seqname not in lSeq:
                lSeq.append( iMap.seqname )
        outFile.write( "%s\t%i\t%i\n" % ( groupID, len(dGr2Mb[groupID]), len(lSeq) ) )
    outFile.close()

    logging.info(string)

#-----------------------------------------------------------------------------

def getStatsForSequences( bsDB, dSeq2Mb, dSeq2Gr, dSeq2Cl, verbose=0 ):

    string = "statistics for sequences:"
    if verbose > 0: print string

    for seq in dSeq2Mb.keys():

        tmpString = "\nseq '%s' (%int) present as %i members in %i groups and %i clusters" % ( seq, bsDB.fetch( seq ).getLength(), len(dSeq2Mb[seq]), len(dSeq2Gr[seq]), len(dSeq2Cl[seq]) )
        
        dBest2Mb = {}
        dBest2Mb[1] = [0,0,0]
        dBest2Mb[2] = [0,0,0]
        
        i = 0
        while i < len(dSeq2Mb[seq]):
            mb = dSeq2Mb[seq][i]
            groupID = mb.name.split("Cl")[0].split("Gr")[1]
            length = mb.length()
            coverage = 100 * int(length) / float(bsDB.fetch( seq ).getLength())
            if coverage >= dBest2Mb[1][2]:
                dBest2Mb[2] = [dBest2Mb[1][0],dBest2Mb[1][1],dBest2Mb[1][2]]
                dBest2Mb[1] = [groupID,length,coverage]
            elif coverage < dBest2Mb[1][2] and coverage > dBest2Mb[2][2]:
                dBest2Mb[2] = [groupID,length,coverage]
            i += 1

        tmpString += "\n1st best group: %s ; %s nt ; %.2f%%" % ( dBest2Mb[1][0], dBest2Mb[1][1], dBest2Mb[1][2] )
        if dBest2Mb[2][0] != 0 and dBest2Mb[2][1] != 0 and dBest2Mb[2][2] != 0 and \
            dBest2Mb[2][0] != dBest2Mb[1][0] and dBest2Mb[2][1] != dBest2Mb[1][1] and dBest2Mb[2][2] != dBest2Mb[1][2]:
            tmpString += "\n2nd best group: %s ; %s nt ; %.2f%%" % ( dBest2Mb[2][0], dBest2Mb[2][1], dBest2Mb[2][2] )

        string += "\n%s" % ( tmpString )
        if verbose > 0:
            print tmpString; sys.stdout.flush()

    string += "\n"
    logging.info(string)

#-----------------------------------------------------------------------------

def getStatsForGroups( bsDB, dGr2Mb, dGr2Seq, dSeq2Gr, dSeq2Cl, verbose=0 ):

    string = "statistics for groups:"
    if verbose > 0: print string

    stat = pyRepet.util.Stat.Stat()

    for groupID in dGr2Mb.keys():
        tmpString1 = "\ngroup %s (%imb,%iseq," % ( groupID, len(dGr2Mb[groupID]), len(dGr2Seq[groupID]) )
        tmpString2 = ""
        stat.reset()
        i = 0
        while i < len(dGr2Mb[ groupID ]):
            mb = dGr2Mb[ groupID ][i]
            if i < 50:
                tmpString2 += " %s(mb%int,seq%int,%igr,%icl)" % ( mb.seqname, mb.length(), bsDB.fetch( mb.seqname ).getLength(), len(dSeq2Gr[mb.seqname]), len(dSeq2Cl[mb.seqname]) )
            stat.add( mb.length() / float( bsDB.fetch( mb.seqname ).getLength() ) )
            i += 1
        if i >= 50:
            tmpString2 += " ..."
        tmpString1 += "%.2f,%.2f):" % ( stat.mean(), stat.cv() )
        string += "\n%s%s" % ( tmpString1, tmpString2 )
        if verbose > 0:
            print tmpString1+tmpString2; sys.stdout.flush()

    string += "\n"
    logging.info(string)

#-----------------------------------------------------------------------------

def getStatsForClusters( dCl2Gr, verbose=0 ):

    string = "statistics for clusters:"
    if verbose > 0: print string

    for clusterID in dCl2Gr.keys():
        tmpString = "cluster %s: %i groups" % ( clusterID, len(dCl2Gr[clusterID]) )
        string += "\n%s" % ( tmpString )
        if verbose > 0:
            print tmpString; sys.stdout.flush()

    string += "\n"
    logging.info(string)

#-----------------------------------------------------------------------------

def main():

    """
    This program gives some details about the clustering results (output from Grouper).
    """

    fastaFileName = ""   #"DmelCaf1_consensus.fa"
    mapFileName = ""   #"DmelCaf1_consensus.align.group.c0.map"
    verbose = 0

    try:
        opts, args = getopt.getopt( sys.argv[1:], "hf:m:v:" )
    except getopt.GetoptError, err:
        print str(err)
        help()
        sys.exit(1)
    for o,a in opts:
        if o == "-h":
            help()
            sys.exit(0)
        elif o == "-f":
            fastaFileName = a
        elif o == "-m":
            mapFileName = a
        elif o == "-v":
            verbose = int(a)

    if fastaFileName == "" or mapFileName == "":
        print "*** Error: missing compulsory options"
        help()
        sys.exit(1)

    if verbose > 0:
        print "\nbeginning of %s" % (sys.argv[0].split("/")[-1])
        sys.stdout.flush()

    logFileName = "giveInfoGrouper.log"
    if os.path.exists(logFileName):
        os.remove(logFileName)
    handler = logging.FileHandler( logFileName )
    formatter = logging.Formatter( "%(asctime)s %(levelname)s: %(message)s" )
    handler.setFormatter( formatter )
    logging.getLogger('').addHandler( handler )
    logging.getLogger('').setLevel( logging.DEBUG )
    logging.info( "started" )

    bsDB = loadInputFastaFile( fastaFileName, verbose )

    dGr2Mb, dGr2Seq, dCl2Mb, dCl2Gr, dSeq2Mb, dSeq2Gr, dSeq2Cl = loadInputMapFile( mapFileName, verbose )

    getSummaryStats( bsDB, mapFileName, dGr2Mb, dGr2Seq, dCl2Mb, dCl2Gr, dSeq2Mb, dSeq2Gr, dSeq2Cl, verbose )

    getStatsForSequences( bsDB, dSeq2Mb, dSeq2Gr, dSeq2Cl, verbose )

    getStatsForGroups( bsDB, dGr2Mb, dGr2Seq, dSeq2Gr, dSeq2Cl, verbose )

    getStatsForClusters( dCl2Gr, verbose )

    logging.info("finished")

    if verbose > 0:
        print "%s finished successfully\n" % (sys.argv[0].split("/")[-1])
        sys.stdout.flush()
        
    return 0

#-----------------------------------------------------------------------------

if __name__ == '__main__':
    main()
