#! /usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import os
import re
import time
import subprocess
from commons.core.seq.BioseqDB import BioseqDB
from commons.core.LoggerFactory import LoggerFactory
from commons.core.checker.CheckerUtils import CheckerUtils
from commons.core.utils.FileUtils import FileUtils
from commons.core.utils.RepetOptionParser import RepetOptionParser
  
LOG_DEPTH = "commons.launcher"
  
class LaunchNucmer(object):

    def __init__(self, queryFileName = "", refFileName = "", prefix = None, genCoords = False, showCoords = False, genVCF = False, mum = False, maxGaps = 90, minMatch = 20, nooptimize = False, mincluster = 65, breaklen = 200, minIdentity = 50, minLength = 100, verbosity = 3):
        self._queryFileName = queryFileName
        self._refFileName = refFileName
        self.setPrefix(prefix)
        self._genCoords = genCoords
        self._showCoords = showCoords
        self._genVCF = genVCF
        self._mum = mum
        self._maxgaps = maxGaps
        self._minMatch = minMatch
        self._nooptimize = nooptimize
        self._mincluster = mincluster
        self._breaklen = breaklen
        
        self._minIdentity = minIdentity
        self._minLength = minLength
        
        self.verbosity = verbosity
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self.verbosity)
        
    def setAttributesFromCmdLine(self):
        usage = "LaunchNucmer.py [options]"
        description = "Launch Nucmer program (part of the mummer package).\n"
        epilog = "Example: launch with info log level, using a specified output prefix and generating a show-coord output.\n"
        epilog += "\t$ LaunchNucmer.py -q refGenome.fa -r queryGenome.fa -p ref_VS_query -s -v 3\n"
        parser = RepetOptionParser(description = description, epilog = epilog, usage = usage)
        parser.add_option("-q", "--query",      dest="queryFileName", default = "",  action="store",    type="string", help="input query file [compulsory] [format: fasta]")
        parser.add_option("-r", "--ref",        dest="refFileName",   default = "",  action="store",    type="string", help="input reference file [compulsory] [format: fasta]")
        parser.add_option("-p", "--prefix",     dest="prefix",        default = None,action="store",    type="string", help="prefix name [optional]")
        parser.add_option("-o", "--gencoords",  dest="genCoords",                    action="store_true",              help="generate coords file with minimal option (show-coords -r) [optional]")
        parser.add_option("-s", "--showcoords", dest="showCoords",                   action="store_true",              help="generate coords file with: show-coords -r -c -l -d -I 50 -L 100 -T [optional]")
        parser.add_option("",   "--vcf",        dest="genVCF",                       action="store_true",              help="generate simple VCF file generated by: show-snps -TCr [optional]")
        parser.add_option("-m", "--mum",        dest="mum",                          action="store_true",              help="use anchor matches that are unique in both the reference and query [optional]")
        parser.add_option("-g", "--maxgaps",    dest="maxgaps",       default = 90,  action="store",    type="int",    help="maximum gap between two adjacent matches in a cluster (default 90) [optional]")
        parser.add_option("-l", "--minmatch",   dest="minMatch",      default = 20,  action="store",    type="int",    help="minimum length of an maximal exact match (default 20) [optional]")
        parser.add_option("-n", "--nooptimize", dest="nooptimize",                   action="store_true",              help="nooptimize (default --optimize) [optional]")
        parser.add_option("-j", "--mincluster", dest="mincluster",    default = 65,  action="store",    type="int",    help="minimum length of a cluster of matches (default 65) [optional]")
        parser.add_option("-b", "--breaklen",   dest="breaklen",      default = 200, action="store",    type="int",    help="Set the distance an alignment extension will attempt to extend poor scoring regions before giving up (default 200) [optional]")
        
        parser.add_option("-i", "--minIdentity",dest="minIdentity",   default = 50,  action="store",    type="int",    help="minimum identity for show-coords (default 50) [optional]")
        parser.add_option("-u", "--minLength",  dest="minLength",     default = 100, action="store",    type="int",    help="minimum alignment length for show-coords (default 100) [optional]")
        parser.add_option("-v", "--verbosity",  dest="verbosity",     default = 3,   action="store",    type="int",    help="verbosity [optional] (default 3)")
        self._options = parser.parse_args()[0]
        self._setAttributesFromOptions(self._options)

    def _setAttributesFromOptions(self, options):
        self._queryFileName = options.queryFileName
        self._refFileName = options.refFileName
        self.setPrefix(options.prefix)
        self._genCoords = options.genCoords
        self._showCoords = options.showCoords
        self._genVCF = options.genVCF
        self._mum = options.mum  
        self._maxgaps = options.maxgaps
        self._minMatch = options.minMatch
        self._nooptimize = options.nooptimize
        self._mincluster = options.mincluster
        self._breaklen = options.breaklen
        
        self._minIdentity = options.minIdentity
        self._minLength = options.minLength
        
        self.verbosity = options.verbosity
        
    def setPrefix(self, prefix):
        if prefix:
            self._prefix = prefix
        else:
            qry = os.path.splitext(os.path.basename(self._queryFileName))[0]
            ref = os.path.splitext(os.path.basename(self._refFileName))[0]
            self._prefix = "%s_VS_%s" % (qry, ref)
            
    def setMincluster(self, value):
        self._mincluster = value
        
    def getMincluster(self):
        return self._mincluster
    
    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise Exception(errorMsg)
            
    def checkOptions(self):       
        if self._queryFileName != "":
            if not FileUtils.isRessourceExists(self._queryFileName):
                self._logAndRaise("ERROR: Query file: %s does not exist!" % self._queryFileName)
        else:
            self._logAndRaise("ERROR: No specified --query option!")
        
        if self._refFileName != "":
            if not FileUtils.isRessourceExists(self._refFileName):
                self._logAndRaise("ERROR: Reference file does not exist!" % self._refFileName)
        else:
            self._logAndRaise("ERROR: No specified --ref option!")
            
    def _convertVARFileIntoVCFFile(self):
        VCFFH = open(self._VCFFile, "w")
        VCFFH.write("##fileformat=VCFv4.1\n")
        VCFFH.write("##fileDate=%s\n" % time.strftime("%Y%m%d"))
        VCFFH.write("##reference=%s\n" % os.path.abspath(self._refFileName))
        VCFFH.write("##assembly=%s\n" % os.path.abspath(self._queryFileName))
        VCFFH.write("##INFO=<ID=SVLEN,Number=.,Type=Integer,Description=\"Difference in length between REF and ALT alleles\">\n")
        VCFFH.write("##INFO=<ID=SVTYPE,Number=1,Type=String,Description=\"Type of structural variant\">\n")
        VCFFH.write("##INFO=<ID=ALTSEQNAME,Number=1,Type=String,Description=\"Name of the sequence where ALT is present\">\n")
        VCFFH.write("##INFO=<ID=ALTSTART,Number=1,Type=Integer,Description=\"ALT start position on query sequence\">\n")
        VCFFH.write("##INFO=<ID=CONTIGCOVERAGE,Number=1,Type=Integer,Description=\"Contig coverage depth\">\n")
        VCFFH.write("##INFO=<ID=SOFTWARE,Number=1,Type=String,Description=\"Software used to generate this VCF\">\n")
        VCFFH.write("#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\n")
        
        iBSDB = BioseqDB(self._refFileName)
        coveragePattern = re.compile("cov_(\d+\.*\d*)")
        with open(self._VARFile, "r") as f:
            line = f.readline()
            variationType = None
            while line:
                isInt = True
                try:
                    int(line[0])
                except ValueError:
                    isInt = False
                
                if isInt:
                    splittedLine = line.split()
                    refSeqname = splittedLine[8]
                    altSeqname = splittedLine[9]
                    contigCov = None
                    allCoverageOccurences = coveragePattern.findall(altSeqname)
                    if len(allCoverageOccurences) == 1:
                        contigCov = float(allCoverageOccurences[0])
                    REF = splittedLine[1]
                    ALT = splittedLine[2]
                    refStart = int(splittedLine[0])
                    altStart = int(splittedLine[3])
                    
                    if REF == "N" or ALT == "N":
                        variationType = "nstretch"
                    elif ALT == ".":
                        variationType = "deletion"
                    elif REF == ".":
                        variationType = "insertion"
                    else:
                        variationType = "snp"
                    
                    if variationType == "deletion":
                        commonNT = iBSDB.fetch(refSeqname).getNtFromPosition(refStart - 1)
                        refSeq = "%s%s" % (commonNT, REF)
                        altSeq = commonNT
                        INFO = "SVTYPE=DEL;AN=2;ALTSEQNAME=%s;ALTSTART=%d;" % (altSeqname, altStart - 1)
                        refLength = 1
                        lineFollow = f.readline()
                        while lineFollow and int(lineFollow.split()[3]) == altStart:
                            refLength += 1
                            refSeq += lineFollow.split()[1]
                            lineFollow = f.readline()
                        
                        INFO += "SVLEN=-%d" % refLength
                        if contigCov:
                            INFO += ";CONTIGCOVERAGE=%f" % contigCov
                        INFO += ";SOFTWARE=LaunchNucmer"
                        VCFLine = "%s\t%d\t.\t%s\t%s\t%s\t%s\t%s\n" % (refSeqname, refStart - 1, refSeq, altSeq, ".", ".", INFO)
                        VCFFH.write(VCFLine)
                        
                    elif variationType == "insertion":
                        commonNT = iBSDB.fetch(refSeqname).getNtFromPosition(refStart - 1)
                        refSeq = commonNT
                        altSeq = "%s%s" % (commonNT, ALT)
                        INFO = "SVTYPE=INS;AN=2;ALTSEQNAME=%s;ALTSTART=%d;" % (altSeqname, altStart - 1)
                        altLength = 1
                        lineFollow = f.readline()
                        while lineFollow and int(lineFollow.split()[0]) == refStart:
                            altLength += 1
                            altSeq += lineFollow.split()[2]
                            lineFollow = f.readline()
                            
                        INFO += "SVLEN=%d" % altLength
                        if contigCov:
                            INFO += ";CONTIGCOVERAGE=%f" % contigCov
                        INFO += ";SOFTWARE=LaunchNucmer"
                        VCFLine = "%s\t%d\t.\t%s\t%s\t%s\t%s\t%s\n" % (refSeqname, refStart - 1, refSeq, altSeq, ".", ".", INFO)
                        VCFFH.write(VCFLine)
                            
                    elif variationType == "snp":
                        INFO = "AN=2;ALTSEQNAME=%s;ALTSTART=%d" % (altSeqname, altStart)
                        if contigCov:
                            INFO += ";CONTIGCOVERAGE=%f" % contigCov
                        INFO += ";SOFTWARE=LaunchNucmer"
                        VCFLine = "%s\t%d\t.\t%s\t%s\t%s\t%s\t%s\n" % (refSeqname, refStart, REF, ALT, ".", ".", INFO)
                        VCFFH.write(VCFLine)
                    
                if variationType == "insertion" or variationType == "deletion":
                    line = lineFollow
                else:
                    line = f.readline()
                                
    def run(self):
        LoggerFactory.setLevel(self._log, self.verbosity)
        self._log.info("START LaunchNucmer")
        if not CheckerUtils.isExecutableInUserPath("nucmer"):
            self._logAndRaise("ERROR: nucmer must be in your path")
        self.checkOptions()
        
        genCoords = ""
        if self._genCoords:
            genCoords = "-o"     
        mum = ""
        if self._mum:
            mum =  "--mum"
        nooptimize = "--optimize"
        if self._nooptimize:
            nooptimize = "--nooptimize" 

        cmd = "nucmer "
        cmd += "%s " % self._refFileName
        cmd += "%s " % self._queryFileName
        cmd += "-p %s " % self._prefix
        cmd += "%s " % genCoords
        cmd += "%s " % mum
        cmd += "-g %d " % self._maxgaps
        cmd += "-l %d " % self._minMatch
        cmd += "-c %d " % self._mincluster
        cmd += "-b %d " % self._breaklen
        cmd += nooptimize
        
        self._log.debug("Running : '%s'" % cmd)
        cmd = cmd.split()
        process = subprocess.Popen(cmd, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
        process.communicate()
        
        if process.returncode != 0:
            self._logAndRaise("ERROR when launching '%s'" % cmd)

        if self._showCoords:
            cmd = "show-coords " 
            cmd += "-r "
            cmd += "-c "
            cmd += "-l "
            cmd += "-d "
            cmd += "-I %d " % self._minIdentity
            cmd += "-L %d " % self._minLength
            cmd += "-T %s.delta > %s.coords" % (self._prefix, self._prefix)
            self._log.debug("Running show-coords with following commands : %s" % cmd)
            os.system(cmd)
            
        if self._genVCF:
            self._VARFile = "%s.var" % self._prefix
            cmd = "show-snps " 
            cmd += "-T "
            cmd += "-C "
            cmd += "-r "
            cmd += "%s.delta > %s" % (self._prefix, self._VARFile)
            self._log.debug("Running show-coords with following commands : %s" % cmd)
            os.system(cmd)
            
            self._VCFFile = "%s.vcf" % self._prefix
            self._log.debug("Converting Nucmer VAR file into VCF file")
            self._convertVARFileIntoVCFFile()
        
        self._log.info("END LaunchNucmer")
        
if __name__ == "__main__":
    iLaunchNucmer = LaunchNucmer()
    iLaunchNucmer.setAttributesFromCmdLine()
    iLaunchNucmer.run()