#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import os
import sys
import time

if not "REPET_PATH" in os.environ.keys():
    print "ERROR: no environment variable REPET_PATH"
    sys.exit(1)
sys.path.append(os.environ["REPET_PATH"])
if not "PYTHONPATH" in os.environ.keys():
    os.environ["PYTHONPATH"] = os.environ["REPET_PATH"]
else:
    os.environ["PYTHONPATH"] = "%s:%s" % (os.environ["REPET_PATH"], os.environ["PYTHONPATH"])

from commons.core.LoggerFactory import LoggerFactory
from commons.core.utils.RepetOptionParser import RepetOptionParser
from commons.core.utils.FileUtils import FileUtils
from commons.core.checker.ConfigChecker import ConfigRules, ConfigChecker
from commons.core.seq.FastaUtils import FastaUtils
from commons.core.seq.BioseqDB import BioseqDB
from commons.core.sql.DbFactory import DbFactory
from commons.core.sql.TableJobAdaptatorFactory import TableJobAdaptatorFactory
from commons.core.launcher.Launcher import Launcher
from commons.core.coord.PathUtils import PathUtils
from commons.core.coord.MergedPath import MergedPath
from commons.core.coord.SplitOnLength import SplitOnLength
from commons.core.stat.Stat import Stat

LOG_DEPTH = "repet.tools"

class LaunchGrouperInParallel(object):
    
    def __init__(self, configFileName = "", fastaFilePath = "", alignFilePath = "", outputFileName = "", groupId = "", doClean = True, verbosity = 3):
        self._configFileName = configFileName
        self.setFastaFilePath(fastaFilePath)
        self.setAlignFilePath(alignFilePath)
        self._outputFileName = outputFileName
        self.setGroupId(groupId)
        self._repetEnvSectionName = "repet_env"
        self._projectSectionName = "project"
        self._prepareBatchesSectionName = "prepare_batches"
        self._grouperSectionName = "cluster_HSPs"      
        self._doClean = doClean
        
        self._lPathFileNames = None
        self._tmpDir = None
        
        self._verbosity = verbosity
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbosity)
    
    def setConfigFileName(self, configFileName):
        self._configFileName = configFileName
    
    def setFastaFilePath(self, fastaFilePath):
        self._fastaFileName = os.path.basename(fastaFilePath)
        
    def setAlignFilePath(self, alignFilePath):
        self._alignFilePath = alignFilePath
        self._alignFileName = os.path.basename(alignFilePath)
    
    def setOutputFileName(self, outputFileName):
        if outputFileName == "":
            self._outputFileName = "%s_Blaster_Grouper_%selem_%sseq.fa" % (self._projectName, self._minNbSeqPerGroup, self._nbLongestSeqPerGroup)
        else:
            self._outputFileName = outputFileName
        
    def setGroupId(self, groupId):
        if groupId == "":
            self._groupId = "Grouper_Parallel_%s" % os.getpid()
        else:
            self._groupId = groupId
        
    def setDoClean(self, doClean):
        self._doClean = doClean
        
    def setVerbosity(self, verbosity):
        self._verbosity = verbosity

    def setAttributesFromCmdLine(self):
        # TODO: improve help usage
        usage = "LaunchGouperInParallel.py [options]"
        description = "Launch Grouper in parallel.\n"
        epilog = "\n"
        parser = RepetOptionParser(description = description, epilog = epilog, usage = usage)
        parser.add_option("-C", "--config",    dest = "configFileName", action = "store", type = "string", help = "configuration file name [compulsory]", default = "")
        parser.add_option("-f", "--fasta",     dest = "fastaFilePath",  action = "store", type = "string", help = "input genome file (format='fasta') [compulsory]", default = "")
        parser.add_option("-a", "--align",     dest = "alignFilePath",  action = "store", type = "string", help = "input align file on genome (format='align') [compulsory]", default = "")
        parser.add_option("-o", "--out",       dest = "outputFileName", action = "store", type = "string", help = "output file name [compulsory]", default = "")
        parser.add_option("-g", "--groupid",   dest = "groupId",        action = "store", type = "string", help = "jobs' group ID to display [optional]", default = "")
        parser.add_option("-v", "--verbosity", dest = "verbosity",      action = "store", type = "int",    help = "verbosity [default: 3]", default = 3)
        options = parser.parse_args()[0]
        self._setAttributesFromOptions(options)
    
    def _setAttributesFromOptions(self, options):
        self.setConfigFileName(options.configFileName)
        self.setFastaFilePath(options.fastaFilePath)
        self.setAlignFilePath(options.alignFilePath)
        self._outputFileName = options.outputFileName
        self.setGroupId(options.groupId)
        self.setVerbosity(options.verbosity)
        
    def _checkOptions(self):
        if self._configFileName == "":
            self._logAndRaise("ERROR: Missing input config file name")
        
        if self._fastaFileName == "":
            self._logAndRaise("ERROR: Missing input fasta file name")
            
        if self._alignFileName == "":
            self._logAndRaise("ERROR: Missing input align file name")
        
    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise Exception(errorMsg)    
    
    def _checkConfig(self):       
        iConfigRules = ConfigRules()
        iConfigRules.addRuleOption(section=self._repetEnvSectionName, option ="repet_host", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._repetEnvSectionName, option ="repet_user", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._repetEnvSectionName, option ="repet_pw", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._repetEnvSectionName, option ="repet_db", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._repetEnvSectionName, option ="repet_port", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._projectSectionName, option ="project_name", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._projectSectionName, option ="project_dir", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._grouperSectionName, option ="resources", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._grouperSectionName, option ="tmpDir", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._grouperSectionName, option ="Grouper_nbGroup", mandatory=True, type="int")
        iConfigRules.addRuleOption(section=self._grouperSectionName, option ="Grouper_coverage", mandatory=True, type="float")
        iConfigRules.addRuleOption(section=self._grouperSectionName, option ="Grouper_include", mandatory=True, type="int")
        iConfigRules.addRuleOption(section=self._grouperSectionName, option ="Grouper_maxJoinLength", mandatory=True, type="int")
        iConfigRules.addRuleOption(section=self._grouperSectionName, option ="minNbSeqPerGroup", mandatory=True, type="int")
        iConfigRules.addRuleOption(section=self._grouperSectionName, option ="nbLongestSeqPerGroup", mandatory=True, type="int")
        iConfigRules.addRuleOption(section=self._grouperSectionName, option ="maxSeqLength", mandatory=True, type="int")
        iConfigRules.addRuleOption(section=self._grouperSectionName, option ="clean", mandatory=True, type="boolean")
        iConfigRules.addRuleOption(section=self._prepareBatchesSectionName, option ="chunk_overlap", mandatory=True, type="int")

        iConfigChecker = ConfigChecker(self._configFileName, iConfigRules)
        self._iConfig = iConfigChecker.getConfig()
        self._setAttributesFromConfig()
        
    def _setAttributesFromConfig(self):
        self._repetHost = self._iConfig.get(self._repetEnvSectionName, "repet_host")
        self._repetUser = self._iConfig.get(self._repetEnvSectionName, "repet_user")
        self._repetPw = self._iConfig.get(self._repetEnvSectionName, "repet_pw")
        self._repetDb = self._iConfig.get(self._repetEnvSectionName, "repet_db")
        self._repetPort = self._iConfig.get(self._repetEnvSectionName, "repet_port")
        self._projectName = self._iConfig.get(self._projectSectionName, "project_name")
        self._projectDir = self._iConfig.get(self._projectSectionName, "project_dir")
        self._resources = self._iConfig.get(self._grouperSectionName, "resources")
        self._tmpDir = self._iConfig.get(self._grouperSectionName, "tmpDir")
        self._nbGrouper = self._iConfig.get(self._grouperSectionName, "Grouper_nbGroup")
        self._coverage = self._iConfig.get(self._grouperSectionName, "Grouper_coverage")
        self._include = self._iConfig.get(self._grouperSectionName, "Grouper_include")
        self._maxJoinLength = self._iConfig.get(self._grouperSectionName, "Grouper_maxJoinLength")
        self._minNbSeqPerGroup = self._iConfig.get(self._grouperSectionName, "minNbSeqPerGroup")
        self._nbLongestSeqPerGroup = self._iConfig.get(self._grouperSectionName, "nbLongestSeqPerGroup")
        self._maxSeqLength = self._iConfig.get(self._grouperSectionName, "maxSeqLength")
        self._doClean = self._iConfig.get(self._grouperSectionName, "clean")
        self._chunkOverlap = self._iConfig.get(self._prepareBatchesSectionName, "chunk_overlap")

    def _getMatcherCmd(self, iLauncher):
        lArgs = []
        lArgs.append("-m %s" % self._alignFileName)
        lArgs.append("-j")
        lArgs.append("-a")
        lArgs.append("-d 10")
        lArgs.append("-g 1")
        lArgs.append("-v %s" % (self._verbosity - 1))
        return iLauncher.getSystemCommand("matcher", lArgs)
    
    def _runMatcher(self):
        cDir = os.getcwd()
        if not self._tmpDir:
            self._tmpDir = cDir
        acronym = "Matcher"
        self._pathFileName = "%s.match.path" % self._alignFileName
        iDb = DbFactory.createInstance()
        iTJA = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
        iLauncher = Launcher(iTJA, cDir, self._tmpDir, self._resources, "%s_%s" % (self._groupId, acronym), acronym)
        
        lCmdStart = []
        lCmdStart.append("os.symlink(\"%s/%s\", \"%s\")" % (cDir, self._alignFileName, self._alignFileName))
        lCmds = []
        lCmds.append(self._getMatcherCmd(iLauncher))
        lCmdFinish = []
        lCmdFinish.append("if os.path.exists(\"%s\"):" % self._pathFileName)
        lCmdFinish.append("\tshutil.move(\"%s\", \"%s/.\")" % (self._pathFileName, cDir))
        
        lCmdsTuples = []
        lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish))
        iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, self._doClean)
        self._lPathFileNames = []
        self._lPathFileNames.append(self._pathFileName)
    
    ## Read the path file (matcher output), create MergedPaths with the Paths and computes the length thresholds
    #
    def _setMergedPathsAndThresholds(self):
        lPaths = PathUtils.getPathListFromFile(self._pathFileName)
        
        dIdToLength = dict((iPath.getIdentifier(), MergedPath()) for iPath in lPaths)
        for path in lPaths:
            dIdToLength[path.getIdentifier()].add(path)

        self._lLengthThresholds = []
        iStats = Stat([iMergedPath.getLength() for iMergedPath in dIdToLength.values()])
        for i in range(1, self._nbGrouper):
            self._lLengthThresholds.append(iStats.quantile(i/float(self._nbGrouper)))
            
        self._lMergedPaths = dIdToLength.values()
    
    ## Split the MergedPaths over the thresholds and create groups of Paths
    #
    def _splitMergedPathsAndCreatePathsGroups(self):
        iSOL = SplitOnLength(self._lMergedPaths, self._lLengthThresholds)
        lMergedPaths_splitOnLength = iSOL.split()
        self._lGroupOfPaths = []
        for lMP in lMergedPaths_splitOnLength:
            lPaths = []
            for MP in lMP:
                lPaths.extend(MP.getPathsList())
            self._lGroupOfPaths.append(lPaths)
    
    ## Write groups of Paths in files to send to Grouper
    #
    def _writeGroupsOfPathsInFiles(self):
        # TODO: Names use [:-5] to remove the original ".path". Should probably be changed.
        self._lPathFileNames = []
        self._lPathFileNames.append("%s_under%s.path" % (self._pathFileName[:-5], str(self._lLengthThresholds[0])))
        if len(self._lLengthThresholds) > 1:
            for item in range(1,len(self._lLengthThresholds)):
                self._lPathFileNames.append("%s_between%s_%s.path" % (self._pathFileName[:-5], str(self._lLengthThresholds[item - 1]), str(self._lLengthThresholds[item])))
        self._lPathFileNames.append("%s_over%s.path" % (self._pathFileName[:-5], str(self._lLengthThresholds[-1])))
        for i in range(0,len(self._lGroupOfPaths)):
            PathUtils.writeListInFile(self._lGroupOfPaths[i], self._lPathFileNames[i], mode="a")
            
    def _runGrouperAndFilterOut(self):
        cDir = os.getcwd()
        if not self._tmpDir:
            self._tmpDir = cDir
        lCmdsTuples = []
        self._lFastaFileNames = []
        for fileName in self._lPathFileNames:
            acronym = "Grouper"
            iDb = DbFactory.createInstance()
            iTJA = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
            iLauncher = Launcher(iTJA, cDir, self._tmpDir, self._resources, "%s_%s" % (self._groupId, acronym))
            lCmds = []
            lCmdStart = []
            lCmdStart.append("os.symlink(\"%s/%s\", \"%s\")" % (cDir, self._fastaFileName, self._fastaFileName))
            lCmdStart.append("os.symlink(\"%s/%s\", \"%s\")" % (cDir, fileName, fileName))
            
            cmd = self._getGrouperCmd(iLauncher, fileName)
            lCmds.append(cmd)
            
            inputFileName = "%s.group.c%s.fa" % (fileName, self._coverage)
            cmd = self._getFilterOutGrouperCmd(iLauncher, inputFileName)
            lCmds.append(cmd)
            
            lCmdFinish = []
            if not self._doClean:
                lCmdFinish.append("shutil.move(\"%s.group.c%s.fa\", \"%s/.\")" % (fileName, self._coverage, cDir))
                lCmdFinish.append("shutil.move(\"%s.group.c%s.param\", \"%s/.\")" % (fileName, self._coverage, cDir))
                lCmdFinish.append("shutil.move(\"%s.group.c%s.set\", \"%s/.\")" % (fileName, self._coverage, cDir))
                lCmdFinish.append("shutil.move(\"%s.group.c%s.txt\", \"%s/.\")" % (fileName, self._coverage, cDir))
                lCmdFinish.append("shutil.move(\"%s.chains.path\",  \"%s/.\")" % (fileName, cDir))
                lCmdFinish.append("shutil.move(\"%s_filtered.log\", \"%s/.\")" % (inputFileName, cDir))
            lCmdFinish.append("shutil.move(\"%s.filtered\", \"%s/.\")" % (inputFileName, cDir))
            lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish))
            self._lFastaFileNames.append("%s.filtered" % inputFileName)
        iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, self._doClean)
        
    def _getGrouperCmd(self, iLauncher, pathFileName):
        lArgs = []
        lArgs.append("-q %s" % self._fastaFileName)
        lArgs.append("-s %s" % self._fastaFileName)
        lArgs.append("-p %s" % pathFileName)
        lArgs.append("-C %s" % self._coverage)
        lArgs.append("-Z %s" % self._minNbSeqPerGroup)
        lArgs.append("-X %s" % self._include)
        lArgs.append("-t 2") #option for grouper multithreading, 2 matches groups in //
#        lArgs.append("-S 3") #option for grouper with 3 sequential matches groups
        lArgs.append("-G -1")
        lArgs.append("-v %i" % (self._verbosity - 2))
        return iLauncher.getSystemCommand("grouper", lArgs)
    
    def _getFilterOutGrouperCmd(self, iLauncher, fastaFileName):
        lArgs = []
        lArgs.append("-i %s" % fastaFileName)
        lArgs.append("-m %s" % self._minNbSeqPerGroup)
        lArgs.append("-M %s" % self._nbLongestSeqPerGroup)
        lArgs.append("-L %s" % self._maxSeqLength)
        lArgs.append("-J %s" % self._maxJoinLength)
        # Note: TODO from the prototype, I don't know what it means
        # TODO: to change when bug fixed with RmvPairAlignInChunkOverlaps.py...
        lArgs.append("-O %s" % self._chunkOverlap)
        lArgs.append("-v %i" % (self._verbosity - 1))
        
        return iLauncher.getSystemCommand("filterOutGrouper.py", lArgs) 
    
    def _getLastClusterNumberFromFastaFile(self, fastaFileName):
        headerList = FastaUtils.dbHeaders(fastaFileName)
        headerNumber = [int(header.split("Cl")[0].split("Gr")[1]) for header in headerList]
        return max(headerNumber)
    
    def _replaceClusterNumberInFastaFromStartAndEnd(self, fastaFileName, start, end):
        iBSDB = BioseqDB(fastaFileName)
        iBSDB.db.sort(key = lambda bs: int(bs.header.split("Gr")[1].split("Cl")[0]), reverse = False)
        count = start
        lastClusterID = 1
        for iBS in iBSDB.db:
            oldHeader = iBS.header
            oldClusterID = int(oldHeader.split("Gr")[1].split("Cl")[0])
            if oldClusterID != lastClusterID:
                count += 1
                
            newHeader = oldHeader.replace("Gr%dCl" % oldClusterID, "Gr%dCl" % count)
            
            lastClusterID = oldClusterID
            iBS.setHeader(newHeader)
            
        with open("%s.renamed" % fastaFileName, "w") as f:
            iBSDB.write(f)
            
    def _makeConsecutiveClusterNumbers(self, listFileNames, outFileName):
        firstFileName = listFileNames[0]
        iBSDB = BioseqDB(firstFileName)
        iBSDB.write(open("%s.renamed" % firstFileName, "w"))
        for f in range(0,len(listFileNames)):
            if f != len(listFileNames)-1:
                maxHeadernb1 = self._getLastClusterNumberFromFastaFile("%s.renamed" % listFileNames[f])
                maxHeadernb2 = self._getLastClusterNumberFromFastaFile(listFileNames[f+1])
                self._replaceClusterNumberInFastaFromStartAndEnd(listFileNames[f+1], maxHeadernb1+1 , maxHeadernb1 + maxHeadernb2)
              
        outputFilesList = ["%s.renamed" % fileName for fileName in listFileNames]
        FileUtils.catFilesFromList(outputFilesList, outFileName, False, False)
        FileUtils.removeFilesByPattern("*.renamed")
    
    def run(self):
        LoggerFactory.setLevel(self._log, self._verbosity)
        self._log.info("START LaunchGrouperInParallel (%s)" % time.strftime("%Y-%m-%d %H:%M:%S"))
        self._checkOptions()
        self._checkConfig()        

        self._log.info("START Matcher (%s)" % time.strftime("%Y-%m-%d %H:%M:%S"))
        self._runMatcher()
        self._log.info("END Matcher (%s)" % time.strftime("%Y-%m-%d %H:%M:%S"))

        if self._nbGrouper > 1:
            self._log.info("START Split (%s)" % time.strftime("%Y-%m-%d %H:%M:%S"))
            self._setMergedPathsAndThresholds()
            self._splitMergedPathsAndCreatePathsGroups()
            self._writeGroupsOfPathsInFiles()
            self._log.info("END Split (%s)" % time.strftime("%Y-%m-%d %H:%M:%S"))
        else:
            self._log.info("No Split (%s)" % time.strftime("%Y-%m-%d %H:%M:%S"))

        self._log.info("START Grouper + FilterOutGrouper (%s)" % time.strftime("%Y-%m-%d %H:%M:%S"))
        self._runGrouperAndFilterOut()
        self._log.info("END Grouper + FilterOutGrouper (%s)" % time.strftime("%Y-%m-%d %H:%M:%S"))

        self.setOutputFileName(self._outputFileName)        
        if(self._nbGrouper > 1):
            self._log.info("START Rename headers + Concatenate (%s)" % time.strftime("%Y-%m-%d %H:%M:%S"))
            self._makeConsecutiveClusterNumbers(self._lFastaFileNames, self._outputFileName)
            self._log.info("END Rename headers + Concatenate (%s)" % time.strftime("%Y-%m-%d %H:%M:%S"))
        else:
            self._log.info("Rename output file (%s)" % time.strftime("%Y-%m-%d %H:%M:%S"))
            os.rename(self._lFastaFileNames[0], self._outputFileName)
        
        if self._doClean:
            FileUtils.removeFilesByPattern("*.path")
            FileUtils.removeFilesByPattern("*.fa.filtered")
        self._log.info("END LaunchGrouperInParallel (%s)" % time.strftime("%Y-%m-%d %H:%M:%S"))
        
    
if __name__ == "__main__":
    iLaunchGrouperInParallel = LaunchGrouperInParallel()
    iLaunchGrouperInParallel.setAttributesFromCmdLine()
    iLaunchGrouperInParallel.run()  