#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import os
import subprocess
from commons.core.LoggerFactory import LoggerFactory
from commons.core.utils.FileUtils import FileUtils
from commons.core.utils.RepetOptionParser import RepetOptionParser
from commons.tools.ProfilesDB4Repet import ProfilesDB4Repet

LOG_DEPTH = "commons.tools"

class HmmProfileFormatter(object):
    
    def __init__(self, inputFileName = "", outFileName = "", superFamily = "", verbosity = 3):
        self._inputFileName = inputFileName
        self.setOutFileName(outFileName)
        self._hmmBuildOutFileName = self._outFileName + ".build"
        self._superFamily = superFamily
        self._verbosity = verbosity
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbosity)
        
    def setAttributesFromCmdLine(self):
        #TODO: Write a tutorial
        usage = "HmmProfileFormatter.py [options]"
        description = "Format a multiple alignment file/ profile etc... into an HMM profile for REPET.\n"
        epilog = "\nExample:\n"
        epilog += "\t$ HmmProfileFormatter.py -i multAlign.fa -v 3\n"
        epilog += "\t Warning : the multiAlign.fa file name must begin with profile name followed with _ \n"
        epilog += "\t if the multiple alignement file corresponds to POL B profile, the name will be POL_profile1\n"
        epilog += "\t if the multiple alignement file corresponds to RNase H profile, the name will be RH_file\n"
        parser = RepetOptionParser(description = description, epilog = epilog, usage = usage)
        parser.add_option("-i", "--fasta",      dest = "fastaFileName", action = "store", type = "string", help = "input multi-aligned fasta file name with extension in {'fa'|'afa'|'fasta'|'fsa'|'fas'} [compulsory] [format: multi-aligned fasta]", default = "")
        parser.add_option("-o", "--out",        dest = "outFileName",   action = "store", type = "string", help = "output file name [default: <input>.hmm]", default = "")
        parser.add_option("-s", "--superfamily",dest = "superFamily",   action = "store", type = "string", help = "name of super family if any", default = "")
        parser.add_option("-v", "--verbosity",  dest = "verbosity",     action = "store", type = "int",    help = "verbosity [optional] [default: 3]", default = 3)
        options = parser.parse_args()[0]
        self._setAttributesFromOptions(options)
        
    def _setAttributesFromOptions(self, options):
        self._inputFileName = options.fastaFileName
        self.setOutFileName(options.outFileName)
        self.sethmmBuildOutFileName(self._outFileName)
        self._superFamily = options.superFamily
        self._verbosity = options.verbosity

    def setOutFileName(self, outFileName):
        if outFileName == "":
            self._outFileName = "%s.hmm" % os.path.splitext(os.path.basename(self._inputFileName))[0]
        else:
            self._outFileName = outFileName
        
    def sethmmBuildOutFileName(self, outFileName):
        self._hmmBuildOutFileName = outFileName + ".build"
        
    def _checkOptions(self):
        if self._inputFileName == "":
            self._logAndRaise("ERROR: Missing input file")
        else:
            if not FileUtils.isRessourceExists(self._inputFileName):
                self._logAndRaise("Input fasta file '%s' does not exist!" % self._inputFileName)
            
    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise Exception(errorMsg)
    
    def _launchHmmBuild(self):
        self._log.info("STARTING hmmbuild")
#TODO see fasta format detection with hmmbuild
        informat = ""
        extension = self._inputFileName.split('.').pop()
        if extension in ["fa","afa","fasta","fsa","fas"]:
            informat =  "--informat afa"
        cmd = "hmmbuild %s %s %s" % (informat, self._hmmBuildOutFileName, self._inputFileName)
#        cmd = "hmmbuild %s %s" % (self._hmmBuildOutFileName, self._inputFileName)
        self._log.debug("Running : '%s'" % cmd)
        process = subprocess.Popen(cmd.split(' '), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        output = process.communicate()
        self._log.debug("Output:\n%s" % output[0])
        if process.returncode != 0:
            self._logAndRaise("ERROR when launching '%s'" % cmd)
            
        self._log.info("FINISHED hmmbuild")
    
    def _formatProfileForRepet(self):
        self._log.info("STARTING ProfilesDB4Repet")
        iProfilesDB4Repet = ProfilesDB4Repet(inputFile = self._hmmBuildOutFileName, outputFile = self._outFileName, superFamily = self._superFamily, verbosity = self._verbosity-1)
        iProfilesDB4Repet.run()
        self._log.info("FINISHED ProfilesDB4Repet")
    
    def run(self):
        LoggerFactory.setLevel(self._log, self._verbosity)
        self._checkOptions()
        self._log.info("START HmmProfileFormatter")
        self._log.debug("Input file name: %s, Output file name: %s, fichier Hmm build %s" % (self._inputFileName,self._outFileName, self._hmmBuildOutFileName ))
        
        self._launchHmmBuild()
        self._formatProfileForRepet()
        
        self._log.info("END HmmProfileFormatter")

if __name__ == "__main__":
    iHmmProfileFormatter = HmmProfileFormatter()
    iHmmProfileFormatter.setAttributesFromCmdLine()
    iHmmProfileFormatter.run()