import os
import sys
import random
import ConfigParser
from decimal import *
from denovo_pipe.GenomeBackbone import GenomeBackbone


class GenomeGrammar(object):
    
    def __init__(self):
        self._verbose = 0
        self._rnd = random
        self._seedForBackboneGeneration = 1859
        self._rnd.seed( self._seedForBackboneGeneration )
        self._lTransitionProbas = []
        
    def setSeedForBackboneGeneration(self, seed):
        self._seed = int(seed)
        self._rnd.seed( self._seed )
        
    def loadConfig(self, configFile):
        config = ConfigParser.ConfigParser()
        config.readfp( open(configFile))
        section = "transition_probabilities"
        if not config.has_section( section ):
            msg = "ERROR: no section '%s' in file '%s'" % ( section, self._configFile )
            sys.stderr.write( "%s\n" % msg )
            sys.exit(1)
        lTransitionProba = []
        for option in config.options( section ):
            iTransProba = TransitionProba( option, config.get(section,option) )
            lTransitionProba.append( iTransProba )
        lTransitionProbaSorted = sorted( lTransitionProba, key=lambda iPair: ( 1 / iPair._proba ) )
        sumProba = 0
        for tp in lTransitionProbaSorted:
            sumProba += tp._proba
        if Decimal(str(sumProba)) != Decimal(1):
            msg = "ERROR: sum of probabilities %f != 1" % ( sumProba )
            sys.stderr.write( "%s\n" % msg )
            sys.exit(1)
        return lTransitionProbaSorted    
        
    def __str__(self):
        msg = ""
        if len(self._lTransitionProbas) > 0:
            msg += "%s" % ( self._lTransitionProbas[0] )
            for tp in self._lTransitionProbas[1:]:
                msg += " %s" % ( tp )
            msg += " seed=%s" % ( self._seed )
        else:
            msg += "seed=%s" % ( self._seed )
        return msg
    
    
    def setTransitionProbas(self, lTransitionProbas):
        self._lTransitionProbas = lTransitionProbas

        
    def makeCumulProbas(self, lTransitionProbas):
        """
        Sort the transition probabilities in descending order.
        Necessary for fitness-proportionate selection.
        @return lFitPropSel
        """
        lFitPropSel = []
        lTransitionProbasSorted = sorted( lTransitionProbas, key=lambda iPair: ( 1 / iPair._proba ) )
        lFitPropSel.append( lTransitionProbasSorted[0]._proba )
        for i in range( 1, len(lTransitionProbasSorted) ):
            lFitPropSel.append( lFitPropSel[-1] + lTransitionProbasSorted[i]._proba )
        return lFitPropSel    
     
     
    def normalizeProbas(self, lTransitionProba):
        sumProba = 0
        lTransitionProbaNormalized = []
        
        for proba in lTransitionProba:
            sumProba += proba._proba
        
        for proba in lTransitionProba:
            lTransitionProbaNormalized.append(TransitionProba(proba._termSymbol, proba._proba / sumProba))
        
        if self._verbose > 1:
            msg = ""
            for proba in lTransitionProbaNormalized:
                msg += "%s " % proba
            print
            print "transition proba normalized %s " % msg
        
        return lTransitionProbaNormalized
    
    
    def randomizeProbas(self, lTransitionProba, seedForRandomizeProbas):
        lTransitionProbaRandomized = []
        
        # be careful to handle the seed properly
        state = self._rnd.getstate()

        self._rnd.seed(seedForRandomizeProbas)
        for iTp in lTransitionProba:
            newTp = TransitionProba( iTp._termSymbol, self._rnd.random() )
            lTransitionProbaRandomized.append( newTp )
        self._rnd.setstate( state )
        
        if self._verbose > 1:
            msg = ""
            for iTp in lTransitionProbaRandomized:
                msg += "%s " % iTp
            print
            print "transition proba randomized %s " % msg
            
        return lTransitionProbaRandomized
    
    def generateGenomeBackbone2(self, seedForBackboneGeneration, transitionProbaConfigFile):
        self._rnd.seed(seedForBackboneGeneration )
        lTransitionProbaSorted = self.loadConfig(transitionProbaConfigFile)
        lFitPropSel = self.makeCumulProbas( lTransitionProbaSorted )
        return self.generateBackbone( lFitPropSel, lTransitionProbaSorted )
    
    def generateBackboneWithSimulations(self, nbTEs, seedForBackboneGeneration , seedForRandomizeProbas):
        if self._verbose > 0:
            print "sample backbones according to conditions..."
            sys.stdout.flush()
            
        lTransitionProba = [ TransitionProba("o", 0),
                            TransitionProba("t", 0),
                            TransitionProba("epsilon", 0) ]
        
        self._rnd.seed(seedForBackboneGeneration )
        iGb = GenomeBackbone()
        
        while True: 
            lTransitionProbaRandomized = self.randomizeProbas( lTransitionProba, seedForRandomizeProbas )
            lTransitionProbaNormalized = self.normalizeProbas( lTransitionProbaRandomized )
            lTransitionProba = lTransitionProbaNormalized
            self.setTransitionProbas( lTransitionProba )
            lFitPropSel = self.makeCumulProbas( lTransitionProba )
            iGb = self.generateBackbone( lFitPropSel, lTransitionProba )
            if iGb.count("t") == nbTEs:
                break
            
        return iGb
 
    def generateBackbone(self, lFitPropSel, lTransitionProbasSorted):
        """
        @param lFitPropSel: fitness-proportionate selection
        """
        if self._verbose > 0:
            print "seed: %i" % ( self._seedForBackboneGeneration )
        iB = GenomeBackbone()
        while True:
            p = self._rnd.random()
            termSymbol = "epsilon"
            for i in range( 0, len(lFitPropSel)-1 ):
                proba = lFitPropSel[i]
                if p <= proba:
                    termSymbol = lTransitionProbasSorted[i]._termSymbol
                    break
            if termSymbol == "epsilon":
                break
            iB.append( termSymbol )
        return iB


class TransitionProba(object):
    
    def __init__(self, termSymbol, proba):
        self._termSymbol = termSymbol
        self._proba = float(proba)
        
    def __str__(self):
        return "%s:%f" % ( self._termSymbol, self._proba )
   
    def __add__(self,o):
        return self._proba + o._proba
        
    
    def __eq__(self, other):
        if type(other) is type(self):
            return self._termSymbol == other._termSymbol and self._proba == other._proba
        return False
    
    def __ne__(self, o):
        return not self.__eq__(o)
